#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Launches steps required to build lignin
Multiple output options, from tcl files to plots
"""
import argparse
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
from collections import (defaultdict)
from configparser import ConfigParser
from common_wrangler.common import (MAIN_SEC, GOOD_RET, INPUT_ERROR, KB, H, KCAL_MOL_TO_J_PART,
                                    INVALID_DATA, OUT_DIR, InvalidDataError, warning, process_cfg, make_dir,
                                    create_out_fname, str_to_file, round_sig_figs)
from rdkit.Chem import (MolToSmiles, MolFromMolBlock)
from rdkit.Chem.AllChem import (Compute2DCoords)
from rdkit.Chem.Draw import MolToFile
from rdkit.Chem.rdMolInterchange import MolToJSON
from ligninkmc import __version__
from ligninkmc.kmc_common import (Event, Monomer, E_BARRIER_KCAL_MOL, E_BARRIER_J_PART, TEMP, INI_MONOS, MAX_MONOS,
                                  SIM_TIME, AFFECTED, GROW, DEF_E_BARRIER_KCAL_MOL, OX, MONOMER, OLIGOMER,
                                  ADJ_MATRIX, RANDOM_SEED, S, G, CHAIN_LEN, BONDS, RCF_YIELDS, RCF_BONDS,
                                  MAX_NUM_DECIMAL, MONO_LIST, CHAIN_MONOS, CHAIN_BRANCH_COEFF, RCF_BRANCH_COEFF,
                                  CHAIN_ID, DEF_CHAIN_ID, PSF_FNAME, DEF_PSF_FNAME, DEF_TOPPAR, TOPPAR_DIR,
                                  MANUSCRIPT_RATES, DEF_RXN_RATES, BOND_TYPE_LIST, INT_TO_TYPE_DICT)
from ligninkmc.kmc_functions import (run_kmc, generate_mol, gen_tcl, count_bonds,
                                     count_oligomer_yields, analyze_adj_matrix)


# Config keys #
CONFIG_KEY = 'config_key'
OUT_FORMAT_LIST = 'output_format_list'
BASENAME = 'outfile_basename'
IMAGE_SIZE = 'image_size'
SAVE_JSON = 'json'
SAVE_PNG = 'png'
SAVE_SMI = 'smi'
SAVE_SVG = 'svg'
SAVE_TCL = 'tcl'
OUT_TYPE_LIST = [SAVE_JSON, SAVE_PNG,  SAVE_SMI, SAVE_SVG, SAVE_TCL]
OUT_TYPE_STR = "', '".join(OUT_TYPE_LIST)
SAVE_FILES = 'save_files_boolean'
ADD_RATES = 'add_rates_list'
RXN_RATES = 'reaction_rates_at_298K'
SG_RATIOS = 'sg_ratio_list'
NUM_REPEATS = 'num_repeats'
DYNAMICS = 'dynamics_flag'
OLIGOMERS = 'oligomers'
MONOMERS = 'monomers'
PLOT_BONDS = 'plot_bonds'
# todo: remove ENERGY_BARRIER_FLAG option when debugging is complete
ENERGY_BARRIER_FLAG = 'energy_barriers_flag'

PLOT_COLORS = [(0, 0, 0), (1, 0, 0), (0, 0, 1), (0, 0.6, 0), (0.6, 0, 0.6), (1, 0.549, 0),
               (0, 0.6, 0.6), (1, 0.8, 0), (0.6078, 0.2980, 0), (0.6, 0, 0), (0, 0, 0.6)]


# Defaults #
DEF_TEMP = 298.15  # K
DEF_MAX_MONOS = 10  # number of monomers
DEF_SIM_TIME = 3600  # simulation time in seconds (1 hour)
DEF_SG = 1
DEF_INI_MONOS = 2
# Estimated addition rate below is based on: https://www.pnas.org/content/early/2019/10/25/1904643116.abstract
#     p. 23121, col2, "0.1 fmol s^-1" for 100micro-m2 surface area; estimated area for lignin modeling is 100nm^2
#     thus 0.1 fmols/ second * 1.00E-15 mol/fmol * 6.022E+23 particles/mol  * 100 nm^2/100microm^2 = 6 monomers/s
#     as an upper limit--rounded down to 1.0 monomers/s--this is just an estimate
DEF_ADD_RATE = 1.0
DEF_IMAGE_SIZE = (1200, 300)
DEF_BASENAME = 'lignin-kmc-out'
DEF_NUM_REPEATS = 1

DEF_VAL = 'default_value'
DEF_CFG_VALS = {OUT_DIR: None, OUT_FORMAT_LIST: None, ADD_RATES: [DEF_ADD_RATE], INI_MONOS: DEF_INI_MONOS,
                MAX_MONOS: DEF_MAX_MONOS, SIM_TIME: DEF_SIM_TIME, SG_RATIOS: [DEF_SG], TEMP: DEF_TEMP,
                RANDOM_SEED: None, BASENAME: DEF_BASENAME, IMAGE_SIZE: DEF_IMAGE_SIZE, DYNAMICS: False,
                E_BARRIER_KCAL_MOL: DEF_E_BARRIER_KCAL_MOL, E_BARRIER_J_PART: None, SAVE_FILES: False,
                SAVE_JSON: False, SAVE_PNG: False, SAVE_SMI: False, SAVE_SVG: False, SAVE_TCL: False,
                CHAIN_ID: DEF_CHAIN_ID, PSF_FNAME: DEF_PSF_FNAME, TOPPAR_DIR: DEF_TOPPAR,
                NUM_REPEATS: DEF_NUM_REPEATS, PLOT_BONDS: False,
                # todo: remove ENERGY_BARRIER_FLAG option when debugging is complete
                ENERGY_BARRIER_FLAG: False,
                }

REQ_KEYS = {}

OPENING_MSG = f"Running Lignin-KMC version {__version__}. " \
              f"Please cite: https://pubs.acs.org/doi/abs/10.1021/acssuschemeng.9b03534\n"


def plot_bond_error_bars(x_axis, y_axis_val_dicts, y_axis_std_dev_dicts, y_val_key_list, x_axis_label, y_axis_label,
                         plot_title, plot_fname):
    plt.figure(figsize=(3.5, 3.5))
    for y_idx, y_key in enumerate(y_val_key_list):
        plt.errorbar(x_axis, y_axis_val_dicts[y_key], yerr=y_axis_std_dev_dicts[y_key], linestyle='none', marker='.',
                     markersize=10, markerfacecolor=PLOT_COLORS[y_idx], markeredgecolor=PLOT_COLORS[y_idx],
                     label=y_key, capsize=3, ecolor=PLOT_COLORS[y_idx])

    if len(x_axis) > 1:
        plt.xscale('log')

    [plt.gca().spines[i].set_linewidth(1.5) for i in ['top', 'right', 'bottom', 'left']]
    plt.gca().tick_params(axis='both', which='major', labelsize=14, direction='in', pad=8, top=True, right=True,
                          width=1.5, length=6)
    plt.gca().tick_params(axis='both', which='minor', labelsize=14, direction='in', pad=8, top=True, right=True,
                          width=1, length=4)
    plt.ylabel(y_axis_label, fontsize=14)
    plt.xlabel(x_axis_label, fontsize=14)

    if '%' in y_axis_label:
        plt.ylim([0.0, 1.0])
    # needed to adjust legends differently for two types of plot
    if y_val_key_list[0] in BOND_TYPE_LIST:
        plt.legend(fontsize=14, loc='upper right', bbox_to_anchor=(1.45, 1.05), frameon=False)
    else:
        plt.legend(fontsize=14, loc='upper right', bbox_to_anchor=(1.7, 1.05), frameon=False)

    plt.title(plot_title)
    plt.savefig(plot_fname, bbox_inches='tight', transparent=True)
    print(f"Wrote file: {plot_fname}")
    plt.close()


def get_avg_percent_bonds(bond_list, num_opts, adj_lists, num_trials):
    """
    Given adj_list for a set of options, with repeats for each option, find the avg and std dev of percent of each
    bond type
    :param bond_list: list of strings representing each bond type
    :param num_opts: number of options specified (should be length of adj_lists)
    :param adj_lists: list of lists of adjs: outer is for each option, inner is for each repeat
    :param num_trials: number of repeats (should be length of inner adj_lists list)
    :return: avg_bonds, std_bonds: list of floats, list of floats: for each option tested, the average and std dev
                  of bond distributions (percentages)
    """
    analysis = []
    for i in range(num_opts):
        cur_adjs = adj_lists[i]
        analysis.append([analyze_adj_matrix(cur_adjs[j]) for j in range(num_trials)])

    bond_percents = {}
    avg_bonds = {}
    std_bonds = {}

    for bond_type in bond_list:
        bond_percents[bond_type] = [[analysis[j][i][BONDS][bond_type]/sum(analysis[j][i][BONDS].values())
                                     for i in range(num_trials)] for j in range(num_opts)]
        avg_bonds[bond_type] = [np.mean(bond_pcts) for bond_pcts in bond_percents[bond_type]]
        std_bonds[bond_type] = [np.sqrt(np.var(bond_pcts)) for bond_pcts in bond_percents[bond_type]]
    return avg_bonds, std_bonds


def create_bond_v_sg_plots(add_rate_str, cfg, sg_adjs):
    all_avg_bonds, all_std_bonds = get_avg_percent_bonds(BOND_TYPE_LIST, len(cfg[SG_RATIOS]), sg_adjs,
                                                         cfg[NUM_REPEATS])
    title = f"Add rate {add_rate_str} monomer/s"
    x_axis_label = 'SG Ratio'
    y_axis_label = 'Bond Type Yield (%)'
    fname = create_out_fname(f'bond_dist_v_sg_{add_rate_str}', base_dir=cfg[OUT_DIR], ext='.png')
    plot_bond_error_bars(cfg[SG_RATIOS], all_avg_bonds, all_std_bonds, BOND_TYPE_LIST,
                         x_axis_label, y_axis_label, title, fname)


def create_dynamics_plots(add_rate_str, bond_types, cfg, num_monos, num_oligs, sg_ratio):
    # Starting with num mon & olig vs timestep:
    len_y_val_key_list = [MONOMERS, OLIGOMERS]
    min_len = len(num_monos[0])
    avg_bond_types = {}
    std_bond_types = {}
    if cfg[NUM_REPEATS] > 1:
        # If there are multiple runs, arrays may be different lengths, so find shortest array
        min_len = len(num_monos[0])
        for mono_list in num_monos[1:]:
            if len(mono_list) < min_len:
                min_len = len(mono_list)
        # make lists of lists into np array
        sg_num_monos = np.asarray([np.array(num_list[:min_len]) for num_list in num_monos])
        # could save; for now, use to make images
        av_num_monos = np.mean(sg_num_monos, axis=0)

        sg_num_oligs = np.asarray([np.array(num_list[:min_len]) for num_list in num_oligs])
        av_num_oligs = np.mean(sg_num_oligs, axis=0)

        std_num_monos = np.std(sg_num_monos, axis=0)
        std_num_oligs = np.std(sg_num_oligs, axis=0)

        len_y_axis_val_dicts = {MONOMERS: av_num_monos, OLIGOMERS: av_num_oligs}
        len_y_axis_std_dev_dicts = {MONOMERS: std_num_monos, OLIGOMERS: std_num_oligs}

        for bond_type in BOND_TYPE_LIST:
            sg_bond_dist = np.asarray([np.array(bond_list[:min_len]) for
                                       bond_list in bond_types[bond_type]])
            avg_bond_types[bond_type] = np.mean(sg_bond_dist, axis=0)
            std_bond_types[bond_type] = np.std(sg_bond_dist, axis=0)

    else:
        len_y_axis_val_dicts = {MONOMERS: num_monos[0], OLIGOMERS: num_oligs[0]}
        len_y_axis_std_dev_dicts = {MONOMERS: None, OLIGOMERS: None}

        for bond_type in BOND_TYPE_LIST:
            avg_bond_types[bond_type] = bond_types[bond_type]
            std_bond_types[bond_type] = None
    timesteps = list(range(min_len))
    title = f"S:G Ratio {sg_ratio}, Add rate {add_rate_str} monomer/s"
    sg_str = f'{sg_ratio:.{3}g}'.replace("+", "").replace(".", "-")
    fname = create_out_fname(f'mono_olig_v_step_{sg_str}_{add_rate_str}', base_dir=cfg[OUT_DIR],
                             ext='.png')
    x_axis_label = 'Time step'
    y_axis_label = 'Number'
    plot_bond_error_bars(timesteps, len_y_axis_val_dicts, len_y_axis_std_dev_dicts, len_y_val_key_list,
                         x_axis_label, y_axis_label, title, fname)
    fname = create_out_fname(f'bond_dist_v_step_{sg_str}_{add_rate_str}', base_dir=cfg[OUT_DIR],
                             ext='.png')
    x_axis_label = 'Time step'
    y_axis_label = 'Number of Bonds'
    plot_bond_error_bars(timesteps, avg_bond_types, std_bond_types, BOND_TYPE_LIST,
                         x_axis_label, y_axis_label, title, fname)


def adj_analysis_to_stdout(adj_results):
    """
    Describe the meaning of the summary dictionary
    :param adj_results: a dictionary from analyze_adj_matrix
    :return: n/a: prints to stdout
    """
    chain_len_results = adj_results[CHAIN_LEN]
    num_monos_created = sum(adj_results[CHAIN_MONOS].values())

    print(f"Lignin KMC created {num_monos_created} monomers, which formed:")
    print_olig_distribution(chain_len_results, adj_results[CHAIN_BRANCH_COEFF])

    lignin_bonds = adj_results[BONDS]
    print(f"composed of the following bond types and number:")
    print_bond_type_num(lignin_bonds)

    print("\nBreaking C-O bonds to simulate RCF results in:")
    print_olig_distribution(adj_results[RCF_YIELDS], adj_results[RCF_BRANCH_COEFF])

    print(f"with the following remaining bond types and number:")
    print_bond_type_num(adj_results[RCF_BONDS])


def print_bond_type_num(lignin_bonds):
    bond_summary = ""
    for bond_type, bond_num in lignin_bonds.items():
        bond_summary += f"   {bond_type.upper():>4}: {bond_num:4}"
    print(bond_summary)


def print_olig_distribution(chain_len_results, coeff):
    for olig_len, olig_num in chain_len_results.items():
        if olig_len == 1:
            print(f"{olig_num:>8} monomer(s) (chain length 1)")
        elif olig_len == 2:
            print(f"{olig_num:>8} dimer(s) (chain length 2)")
        elif olig_len == 3:
            print(f"{olig_num:>8} trimer(s) (chain length 3)")
        else:
            print(f"{olig_num:>8} oligomer(s) of chain length {olig_len}, with branching coefficient "
                  f"{round(coeff[olig_len], 3)}")


def degree(adj):
    """
    Determines the degree for each monomer within the polymer chain. The "degree" concept in graph theory
    is the number of edges connected to a node. In the context of lignin, that is simply the number of
    connected residues to a specific residue, and can be used to determine derived properties like the
    branching coefficient.
    :param adj: scipy dok_matrix   -- the adjacency matrix for the lignin polymer that has been simulated
    :return: The degree for each monomer as a numpy array.
    """
    return np.bincount(adj.nonzero()[0])


def overall_branching_coefficient(adj):
    """
    Based on the definition in Dellon et al. (10.1021/acs.energyfuels.7b01150), this is the number of
       branched oligomers divided by the total number of monomers.
    This value is indifferent to the number of fragments in the output.

    :param adj: dok_matrix, the adjacency matrix for the lignin polymer that has been simulated
    :return: The branching coefficient that corresponds to the adjacency matrix
    """
    degrees = degree(adj)
    if len(degrees) == 0:
        return 0
    else:
        return np.sum(degrees >= 3) / len(degrees)


def get_bond_type_v_time_dict(adj_list, sum_len_larger_than=None):
    """
    given a list of adjs (one per timestep), flip nesting so have dictionaries of lists of type val vs. time
    for graphing, also created a 10+ list
    :param adj_list: list of adj dok_matrices
    :param sum_len_larger_than: None or an integer; if an integer, make a val_list that sums all lens >= that value
    :return: two dict of dicts
    """
    bond_type_dict = defaultdict(list)
    # a little more work for olig_len_monos_dict, since each timestep does not contain all possible keys
    olig_len_monos_dict = defaultdict(list)
    olig_len_count_dict = defaultdict(list)
    olig_count_dict_list = []
    frag_count_dict_list = []  # first make list of dicts to get max bond_length
    for adj in adj_list:  # loop over each timestep
        # this is keys = timestep  values
        count_bonds_list = count_bonds(adj)
        for bond_type in count_bonds_list:
            bond_type_dict[bond_type].append(count_bonds_list[bond_type])
        olig_yield_dict, olig_monos_dict, olig_branch_dict, olig_branch_coeff_dict = count_oligomer_yields(adj)
        olig_count_dict_list.append(olig_yield_dict)
        frag_count_dict_list.append(olig_monos_dict)
    # since breaking bonds is not allowed, the longest oligomer will be from the last step; ordered, so last len
    max_olig_len = list(frag_count_dict_list[-1].keys())[-1]
    # can now get the dict of lists from list of dicts
    for frag_count_list, olig_count_list in zip(frag_count_dict_list, olig_count_dict_list):
        for olig_len in range(1, max_olig_len + 1):
            olig_len_monos_dict[olig_len].append(frag_count_list.get(olig_len, 0))
            olig_len_count_dict[olig_len].append(olig_count_list.get(olig_len, 0))
    # now make a list of all values greater than a number, if given
    # first initialize as None so something can be returned, even if we are not summing over a particular number
    len_monos_plus_list = None
    len_count_plus_list = None
    if sum_len_larger_than:
        num_time_steps = len(adj_list)
        len_monos_plus_list = np.zeros(num_time_steps)
        len_count_plus_list = np.zeros(num_time_steps)
        # both dicts have same keys, so no worries
        for olig_len, val_list in olig_len_monos_dict.items():
            if olig_len >= sum_len_larger_than:
                len_monos_plus_list = np.add(len_monos_plus_list, val_list)
                len_count_plus_list = np.add(len_count_plus_list, olig_len_count_dict[olig_len])
    return bond_type_dict, olig_len_monos_dict, len_monos_plus_list, olig_len_count_dict, len_count_plus_list


def read_cfg(f_loc, cfg_proc=process_cfg):
    """
    Reads the given configuration file, returning a dict with the converted values supplemented by default values.

    :param f_loc: The location of the file to read.
    :param cfg_proc: The processor to use for the raw configuration values.  Uses default values when the raw
        value is missing.
    :return: A dict of the processed configuration file's data.
    """
    config = ConfigParser()
    good_files = config.read(f_loc)

    if not good_files:
        raise IOError(f"Could not find specified configuration file: {f_loc}")

    main_proc = cfg_proc(dict(config.items(MAIN_SEC)), DEF_CFG_VALS, REQ_KEYS)

    return main_proc


def parse_cmdline(argv=None):
    """
    Returns the parsed argument list and return code.
    :param argv: A list of arguments, or `None` for ``sys.argv[1:]``.
    """
    if argv is None:
        argv = sys.argv[1:]

    # initialize the parser object:
    parser = argparse.ArgumentParser(description=f"Create lignin chain(s) composed of 'S' ({S}) and/or 'G' ({G}) "
                                                 f"monolignols, as described in:\n  Orella, M., "
                                                 f"Gani, T. Z. H., Vermaas, J. V., Stone, M. L., Anderson, E. M., "
                                                 f"Beckham, G. T., \n  Brushett, Fikile R., Roman-Leshkov, Y. (2019). "
                                                 f"Lignin-KMC: A Toolkit for Simulating Lignin Biosynthesis.\n  "
                                                 f"ACS Sustainable Chemistry & Engineering. "
                                                 f"https://doi.org/10.1021/acssuschemeng.9b03534. C-Lignin can be \n  "
                                                 f"modeled with the functions in this package, as shown in ipynb "
                                                 f"examples in our project package on github \n  "
                                                 f"(https://github.com/michaelorella/lignin-kmc/), but not currently "
                                                 f"from the command line. If this \n  functionality is desired, "
                                                 f"please start a new issue on the github.\n\n  By default, the Gibbs "
                                                 f"free energy barriers from this reference will be used, as specified "
                                                 f"in Tables S1 and S2.\n  Alternately, the user may specify values, "
                                                 f"which should be specified as a dict of dict of dicts in a \n  "
                                                 f"specified configuration file (specified with '-c') using the "
                                                 f"'{E_BARRIER_KCAL_MOL}' or '{E_BARRIER_J_PART}'\n  parameters with "
                                                 f"corresponding units (kcal/mol or joules/particle, respectively), in "
                                                 f"a configuration file \n  (see '-c'). The format is (bond_type: "
                                                 f"monomer(s) involved: units involved: ea_vals), for example:\n      "
                                                 f"ea_dict = {{{OX}: {{'G': {{{MONOMER}: 0.9, {OLIGOMER}: 6.3}}, "
                                                 f"'S': ""{{{MONOMER}: 0.6, {OLIGOMER}: " f"2.2}}}}, ...}}.\n  "
                                                 f"The default output is a SMILES string printed to standard out.\n\n"
                                                 f"  All command-line options may alternatively be specified in a "
                                                 f"configuration file. Command-line (non-default) \n  selections will "
                                                 f"override configuration file specifications.",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("-a", "--add_rates", help=f"A comma-separated list of the rates of monomer addition to the "
                                                  f"system (in monomers/second), \nto be used when the '{MAX_MONOS}' "
                                                  f"('-m' option) is larger than '{INI_MONOS}' \n('-i' option), thus "
                                                  f"specifying monomer addition. The simulation will end when either "
                                                  f"there \nare no more possible reactions (including monomer "
                                                  f"addition) or when the '{SIM_TIME}' \n('-l' option) is reached, "
                                                  f"whichever comes first. Note: if there are spaces in the list of "
                                                  f"\naddition rates, the list must be enclosed in quotes to be read "
                                                  f"as a single string. The \ndefault list contains the single "
                                                  f"addition rate of {DEF_ADD_RATE} monomers/s.",
                        default=[DEF_ADD_RATE])
    parser.add_argument("-c", "--config", help="The location of the configuration file in the 'ini' format. This file "
                                               "can be used to \noverwrite default values such as for energies.",
                        default=None, type=read_cfg)
    parser.add_argument("-d", "--out_dir", help="The directory where output files will be saved. The default is "
                                                "the current directory.", default=DEF_CFG_VALS[OUT_DIR])
    parser.add_argument("-dy", "--dynamics_flag", help=f"Select this option if dynamics (results per timestep) are "
                                                       f"requested. If chosen, plots of \nmonomers and oligomers vs "
                                                       f"timestep, and bond type percent vs timestep, will be saved. "
                                                       f"\nThey will be named 'bond_dist_v_step_*_#.png' and "
                                                       f"'mono_olig_v_step_*_#.png', where * \nrepresents the S:G "
                                                       f"ratio and # represents the addition rate. Note that this "
                                                       f"option \nsignificantly increases simulation time.",
                        action="store_true")
    # todo: remove -e option when debugging is complete
    parser.add_argument("-e", "--energy_barriers_flag", help=f"By default, the reaction rates will be based on the "
                                                             f"energy barriers in kcal/mol to be used \nto calculate "
                                                             f"reaction rates at {DEF_TEMP} K. If this flag is used, "
                                                             f"the rates use to produce the \noriginal manuscript "
                                                             f"plots will be used (missing one term). Alternate sets "
                                                             f"of energy \nbarriers can be specified; see the main "
                                                             f"help message.", action="store_true")
    parser.add_argument("-f", "--output_format_list", help="The type(s) of output format to be saved. Provide as a "
                                                           "space- or comma-separated list. \nNote: if the list has "
                                                           "spaces, it must be enclosed in quotes, to be treated as "
                                                           "a single \nstring. The currently supported "
                                                           f"types are: '{OUT_TYPE_STR}'. \nThe '{SAVE_JSON}' "
                                                           f"option will save a json format of RDKit's 'mol' "
                                                           f"(molecule) object. The '{SAVE_TCL}' \noption will create "
                                                           f"a file for use with VMD to generate a psf file and 3D "
                                                           f"molecules, \nas described in LigninBuilder, "
                                                           f"https://github.com/jvermaas/LigninBuilder, \n"
                                                           f"https://pubs.acs.org/doi/abs/10.1021/acssuschemeng.8b05665"
                                                           f". \nA base name for the saved "
                                                           f"files can be provided with the '-o' option. Otherwise, "
                                                           f"the \nbase name will be '{DEF_BASENAME}'.",
                        default=DEF_CFG_VALS[OUT_FORMAT_LIST])
    parser.add_argument("-i", "--initial_num_monomers", help=f"The initial number of monomers to be included in the "
                                                             f"simulation. The default is {DEF_INI_MONOS}.",
                        default=DEF_CFG_VALS[INI_MONOS])
    parser.add_argument("-l", "--length_simulation", help=f"The length of simulation (simulation time) in seconds. The "
                                                          f"default is {DEF_SIM_TIME} s.", default=DEF_SIM_TIME)
    parser.add_argument("-m", "--max_num_monomers", help=f"The maximum number of monomers to be studied. The default "
                                                         f"value is {DEF_MAX_MONOS}.", default=DEF_MAX_MONOS)
    parser.add_argument("-n", "--num_repeats", help=f"The number of times each combination of sg_ratio and add_rate "
                                                    f"will be tested. The default is {DEF_NUM_REPEATS}.",
                        default=DEF_NUM_REPEATS)
    parser.add_argument("-o", "--output_basename", help=f"The base name for output file(s). If an extension is "
                                                        f"provided, it will determine \nthe type of output. Currently "
                                                        f"supported output types are: \n'{OUT_TYPE_STR}'. Multiple "
                                                        f"output formats can be selected with the \n'-f' option. If "
                                                        f"the '-f' option is selected and no output base name "
                                                        f"provided, a default \nbase name of '{DEF_BASENAME}' will be "
                                                        f"used.", default=DEF_BASENAME)
    parser.add_argument("-p", "--plot_bonds", help=f"Flag to produce plots of the percent of each bond type versus S:G "
                                                   f"ratio(s). One plot will \nbe created per addition rate, named "
                                                   f"'bond_dist_v_sg_#.png', where # represents \nthe addition rate.",
                        action="store_true")
    parser.add_argument("-r", "--random_seed", help="A positive integer to be used as a seed value for testing. The "
                                                    "default is not to use a \nseed, to allow pseudorandom lignin "
                                                    "creation.", default=DEF_CFG_VALS[RANDOM_SEED])
    parser.add_argument("-s", "--image_size", help=f"The output size of svg or png files in pixels. The default size "
                                                   f"is {DEF_IMAGE_SIZE} pixels. \nTo use a different size, provide "
                                                   f"two integers, separated by a space or a comma. \nNote: if the "
                                                   f"list of two numbers has any spaces in it, it must be enclosed "
                                                   f"in quotes.",
                        default=DEF_IMAGE_SIZE)
    parser.add_argument("-sg", "--sg_ratios", help=f"A comma-separated list of the S:G (guaiacol:syringyl) ratios to "
                                                   f"be tested. \nIf there are spaces, the list must be enclosed in "
                                                   f"quotes to be read as a single string. \nThe default list "
                                                   f"contains the single value {DEF_SG}.", default=[DEF_SG])
    parser.add_argument("-t", "--temperature_in_k", help=f"The temperature (in K) at which to model lignin "
                                                         f"biosynthesis. The default is {DEF_TEMP} K.\nNote: this "
                                                         f"temperature must match the temperature at which the "
                                                         f"energy barriers were calculated. ",
                        default=DEF_TEMP)
    parser.add_argument("--chain_id", help=f"Option for use when generating a tcl script: the chainID to be used in "
                                           f"generating a psf \nand/or pdb file from a tcl script (see LigninBuilder). "
                                           f"This should be one character. If a \nlonger ID is provided, it will be "
                                           f"truncated to the first character. The default value is {DEF_CHAIN_ID}.",
                        default=DEF_CHAIN_ID)
    parser.add_argument("--psf_fname", help=f"Option for use when generating a tcl script: the file name for psf and "
                                            f"pdb files that will \nbe produced from running a tcl produced by this "
                                            f"package (see LigninBuilder). The default \nvalue is {DEF_PSF_FNAME}.",
                        default=DEF_PSF_FNAME)
    parser.add_argument("--toppar_dir", help=f"Option for use when generating a tcl script: the directory name where "
                                             f"VMD should look for \nthe toppar file(s) when running the tcl file in "
                                             f"VMD (see LigninBuilder). The default value \nis '{DEF_TOPPAR}'.",
                        default=DEF_TOPPAR)

    args = None
    try:
        args = parser.parse_args(argv)
        # dict below to map config input and defaults to command-line input
        conf_arg_dict = {OUT_DIR: args.out_dir,
                         OUT_FORMAT_LIST: args.output_format_list,
                         ADD_RATES: args.add_rates,
                         DYNAMICS: args.dynamics_flag,
                         INI_MONOS: args.initial_num_monomers,
                         SIM_TIME: args.length_simulation,
                         MAX_MONOS: args.max_num_monomers,
                         BASENAME: args.output_basename,
                         IMAGE_SIZE: args.image_size,
                         SG_RATIOS: args.sg_ratios,
                         TEMP: args.temperature_in_k,
                         RANDOM_SEED: args.random_seed,
                         CHAIN_ID: args.chain_id,
                         PSF_FNAME: args.psf_fname,
                         TOPPAR_DIR: args.toppar_dir,
                         NUM_REPEATS: args.num_repeats,
                         PLOT_BONDS: args.plot_bonds,
                         # todo: remove -e option when debugging is complete
                         ENERGY_BARRIER_FLAG: args.energy_barriers_flag,
                         }
        if args.config is None:
            args.config = DEF_CFG_VALS.copy()
        # Now overwrite any config values with command-line arguments, only if those values are not the default
        for config_key, arg_val in conf_arg_dict.items():
            if not (arg_val == DEF_CFG_VALS[config_key]):
                args.config[config_key] = arg_val

    except (KeyError, IOError, SystemExit) as e:
        if hasattr(e, 'code') and e.code == 0:
            return args, GOOD_RET

        # only print the e if it has meaningful info; 2 simply is system exit from parser;
        #    tests that have triggered System Exit are caught and explained below
        if not e.args[0] == 2:
            warning(e)

        # Easy possible error is to have a space in a list; check for it
        check_arg_list = []

        for arg_str in ['-f', '--output_format_list', '-s', '--image_size', "-a", "--add_rates", "-sg", "--sg_ratio"]:
            if arg_str in argv:
                check_arg_list.append(arg_str)
        if len(check_arg_list) > 0:
            check_list = "', '".join(check_arg_list)
            warning(f"Check your entry/entries for '{check_list}'. If spaces separate list entries, "
                    f"enclose the whole list in quotes, or separate with commas only.")

        parser.print_help()
        return args, INPUT_ERROR
    return args, GOOD_RET


def calc_rates(temp, ea_j_part_dict=None, ea_kcal_mol_dict=None):
    """
    Uses temperature and provided Gibbs free energy barriers (at 298.15 K and 1 atm) to calculate rates using the
        Eyring equation
    Dict formats: dict = {rxn_type: {substrate(s): {sub_lengths (e.g. (monomer, monomer)): value, ...}, ...}, ...}

    Only ea_j_part_dict or ea_kcal_mol_dict are needed; if both are provided, only ea_j_part_dict will be used

    :param temp: float, temperature in K
    :param ea_j_part_dict: dictionary of Gibbs free energy barriers in Joule/particle, in format noted above
    :param ea_kcal_mol_dict: dictionary of Gibbs free energy barriers in kcal/mol, in format noted above
    :return: rxn_rates: dict of reaction rates units of 1/s
    """
    # want Gibbs free energy barriers in J/particle for later calculation;
    #     user can provide them in those units or in kcal/mol
    if ea_j_part_dict is None:
        ea_j_part_dict = {
            rxn_type: {substrate: {sub_len: ea_kcal_mol_dict[rxn_type][substrate][sub_len] * KCAL_MOL_TO_J_PART
                                   for sub_len in ea_kcal_mol_dict[rxn_type][substrate]}
                       for substrate in ea_kcal_mol_dict[rxn_type]} for rxn_type in ea_kcal_mol_dict}
    rxn_rates = {}
    for rxn_type in ea_j_part_dict:
        rxn_rates[rxn_type] = {}
        for substrate in ea_j_part_dict[rxn_type]:
            rxn_rates[rxn_type][substrate] = {}
            for substrate_type in ea_j_part_dict[rxn_type][substrate]:
                # rounding to reduce difference due to solely to platform running package
                rate = KB * temp / H * np.exp(-ea_j_part_dict[rxn_type][substrate][substrate_type] / KB / temp)
                rxn_rates[rxn_type][substrate][substrate_type] = round_sig_figs(rate, sig_figs=15)
    return rxn_rates


def create_initial_monomers(pct_s, monomer_draw):
    """
    Make a monomer list (length of monomer_draw) based on the types determined by the monomer_draw list and pct_s
    :param pct_s: float ([0:1]), fraction of  monomers that should be type "S"
    :param monomer_draw: a list of floats ([0:1)) to determine if the monomer should be type "G" (val < pct_s) or
                         "S", otherwise
    :return: list of Monomer objects of specified type
    """
    # TODO: If want more than 2 monomer options, need to change logic; that will require an overhaul, since
    #       sg_ratio is often used. However, until we have Gibbs free energy barriers for bonding between more than
    #       just S and G, no need to update
    # if mon_choice < pct_s, make it an S; that is, the evaluation comes back True (=1='S');
    #     otherwise, get False = 0 = 'G'. Since only two options (True/False) only works for 2 monomers
    return [Monomer(INT_TO_TYPE_DICT[int(mono_type_draw < pct_s)], i) for i, mono_type_draw in enumerate(monomer_draw)]


def create_initial_events(initial_monomers, rxn_rates):
    """
    # Create event_dict that will oxidize every monomer
    :param initial_monomers: a list of Monomer objects
    :param rxn_rates: dict of dict of dicts of reaction rates in 1/s
    :return: a list of oxidation Event objects to initialize the state by allowing oxidation of every monomer
    """
    return [Event(OX, [mon.identity], rxn_rates[OX][mon.type][MONOMER]) for mon in initial_monomers]


def create_initial_state(initial_events, initial_monomers):
    return {i: {MONOMER: initial_monomers[i], AFFECTED: {initial_events[i]}} for i in range(len(initial_monomers))}


def produce_output(adj_matrix, mono_list, cfg):
    # Default out is SMILES
    block = generate_mol(adj_matrix, mono_list)
    mol = MolFromMolBlock(block)
    smi_str = MolToSmiles(mol) + '\n'
    # if SMI is to be saved, don't output to stdout
    if cfg[SAVE_SMI]:
        fname = create_out_fname(cfg[BASENAME], base_dir=cfg[OUT_DIR], ext=SAVE_SMI)
        str_to_file(smi_str, fname, print_info=True)
    else:
        print("\nSMILES representation: \n", MolToSmiles(mol), "\n")
    if cfg[SAVE_PNG] or cfg[SAVE_SVG] or cfg[SAVE_JSON]:
        # PNG and SVG make 2D images and thus need coordinates
        # JSON will save coordinates--zero's if not computed; might as well compute and save non-zero values
        Compute2DCoords(mol)
    for save_format in [SAVE_TCL, SAVE_JSON, SAVE_PNG, SAVE_SVG]:
        if cfg[save_format]:
            fname = create_out_fname(cfg[BASENAME], base_dir=cfg[OUT_DIR], ext=save_format)
            if save_format == SAVE_TCL:
                gen_tcl(adj_matrix, mono_list, tcl_fname=fname, chain_id=cfg[CHAIN_ID],
                        psf_fname=cfg[PSF_FNAME], toppar_dir=cfg[TOPPAR_DIR], out_dir=cfg[OUT_DIR])
            if save_format == SAVE_JSON:
                json_str = MolToJSON(mol)
                str_to_file(json_str + '\n', fname)
            elif save_format == SAVE_PNG or save_format == SAVE_SVG:
                MolToFile(mol, fname, size=cfg[IMAGE_SIZE])
            print(f"Wrote file: {fname}")


def initiate_state(add_rate, cfg, rep, sg_ratio):
    pct_s = sg_ratio / (1 + sg_ratio)
    ini_num_monos = cfg[INI_MONOS]
    if cfg[RANDOM_SEED]:
        # we don't want the same random seed for every iteration
        np.random.seed(cfg[RANDOM_SEED] + int(add_rate / 100 + sg_ratio * 10) + rep)
        monomer_draw = np.around(np.random.rand(ini_num_monos), MAX_NUM_DECIMAL)
    else:
        monomer_draw = np.random.rand(ini_num_monos)
    initial_monomers = create_initial_monomers(pct_s, monomer_draw)
    # initial event must be oxidation to create reactive species; all monomers may be oxidized
    initial_events = create_initial_events(initial_monomers, cfg[RXN_RATES])
    # initial_monomers and initial_events are grouped into the initial state
    initial_state = create_initial_state(initial_events, initial_monomers)
    if cfg[MAX_MONOS] > cfg[INI_MONOS]:
        initial_events.append(Event(GROW, [], rate=add_rate))
    elif cfg[MAX_MONOS] < cfg[INI_MONOS]:
        warning(f"The specified maximum number of monomers ({cfg[MAX_MONOS]}) is less than the "
                f"specified initial number of monomers ({cfg[INI_MONOS]}). \n The program will "
                f"proceed with the with the initial number of monomers with no addition of monomers.")
    return initial_events, initial_state


def validate_input(cfg):
    """
    Checking for errors at the beginning, so don't waste time starting calculations that will not be able to complete

    :param cfg: dict of configuration values
    :return: will raise an error if invalid data is encountered
    """
    # Don't use "if cfg[RANDOM_SEED]:", because that won't catch the user giving the value 0, which they might think
    #    would be a valid random seed, but won't work for this package because of later "if cfg[RANDOM_SEED]:" checks
    if cfg[RANDOM_SEED] is not None:
        try:
            # numpy seeds must be 0 and 2**32 - 1. Raise an error if the input cannot be converted to an int. Also raise
            #   an error for 0, since that will return False that a seed was provided in the logic in this package
            cfg[RANDOM_SEED] = int(cfg[RANDOM_SEED])
            if cfg[RANDOM_SEED] <= 0 or cfg[RANDOM_SEED] > (2**32 - 1):
                raise ValueError
        except ValueError:
            raise InvalidDataError(f"Invalid input provided for '{RANDOM_SEED}': '{cfg[RANDOM_SEED]}'. If you "
                                   f"would like to obtain consistent output by using a random seed, provide a "
                                   f"positive integer value no greater than 2**32 - 1.")

    # Convert list entries. Will already be lists if defaults are used. Otherwise, they will be strings.
    list_args = [ADD_RATES, SG_RATIOS]
    arg, arg_val = "", ""  # to make IDE happy
    try:
        for arg in list_args:
            arg_val = cfg[arg]
            # Will be a string to process unless it is the default
            if isinstance(arg_val, str):
                raw_vals = arg_val.replace(",", " ").replace("(", "").replace(")", "").split()
                cfg[arg] = [float(val) for val in raw_vals]
            else:
                cfg[arg] = arg_val
            for val in cfg[arg]:
                if val < 0:
                    raise ValueError
                # okay for sg_ratio to be zero, but not add_rate
                elif val == 0 and arg == ADD_RATES:
                    raise ValueError
    except ValueError:
        raise InvalidDataError(f"Found {arg_val} for '{arg}'. This entry must be able to be "
                               f"converted to a list of positive floats.")

    # now testing for positive floats
    for req_pos_num in [SIM_TIME]:
        try:
            cfg[req_pos_num] = float(cfg[req_pos_num])
            if cfg[req_pos_num] <= 0:
                raise ValueError
        except ValueError:
            raise InvalidDataError(f"Found '{cfg[req_pos_num]}' input for '{req_pos_num}'. The {req_pos_num} must be "
                                   f"a positive number.")

    # Required ints
    for req_pos_int_arg in [INI_MONOS, MAX_MONOS, NUM_REPEATS]:
        try:
            cfg[req_pos_int_arg] = int(cfg[req_pos_int_arg])
            if cfg[req_pos_int_arg] < 0:
                raise ValueError
        except ValueError:
            raise InvalidDataError(f"Found '{cfg[req_pos_int_arg]}' input for '{req_pos_int_arg}'. The "
                                   f"{req_pos_int_arg} must be a positive integer.")

    try:
        # Will be a string to process unless it is the default
        if isinstance(cfg[IMAGE_SIZE], str):
            raw_vals = cfg[IMAGE_SIZE].replace(",", " ").replace("(", "").replace(")", "").split()
            if len(raw_vals) != 2:
                raise ValueError
            cfg[IMAGE_SIZE] = (int(raw_vals[0]), int(raw_vals[1]))
    except ValueError:
        raise InvalidDataError(f"Found '{cfg[IMAGE_SIZE]}' input for '{IMAGE_SIZE}'. The {IMAGE_SIZE} must be "
                               f"two positive numbers, separated either by a comma or a space.")

    # Check for valid output requests
    check_if_files_to_be_saved(cfg)

    # determine rates to use
    if cfg[E_BARRIER_KCAL_MOL] == DEF_CFG_VALS[E_BARRIER_KCAL_MOL] and (cfg[E_BARRIER_J_PART] ==
                                                                        DEF_CFG_VALS[E_BARRIER_J_PART]
                                                                        ) and (cfg[TEMP] == DEF_TEMP):
        # todo: remove ENERGY_BARRIER_FLAG option when debugging is complete
        if cfg[ENERGY_BARRIER_FLAG]:
            cfg[RXN_RATES] = MANUSCRIPT_RATES
        else:
            cfg[RXN_RATES] = DEF_RXN_RATES
    else:
        # todo: remove ENERGY_BARRIER_FLAG option when debugging is complete
        if cfg[ENERGY_BARRIER_FLAG]:
            warning("Both the {ENERGY_BARRIER_FLAG} option and energy barriers have been provided. The "
                    "{ENERGY_BARRIER_FLAG} option will be ignored, and reaction rates will be calculated from the "
                    "provided energy barriers.")
        if int(cfg[TEMP]) != int(DEF_TEMP):
            warning(f"The program will continue, using a temperature other than {DEF_TEMP}. Ensure that the energy "
                    f"barriers being used where calculated at the provided temperature ({cfg[TEMP]}), otherwise "
                    f"cancel this run.")
        cfg[RXN_RATES] = calc_rates(cfg[TEMP], ea_j_part_dict=cfg[E_BARRIER_J_PART],
                                    ea_kcal_mol_dict=cfg[E_BARRIER_KCAL_MOL])


def check_if_files_to_be_saved(cfg):
    """
    Evaluate input for requests to save output and check for valid specified locations
    :param cfg: dict of configuration values
    :return: if the cfg designs that files should be created, returns an updated cfg dict, and raises errors if
              invalid data in encountered
    """
    if cfg[OUT_FORMAT_LIST]:
        # remove any periods to aid comparison; might as well also change comma to space and then split on just space
        out_format_list = cfg[OUT_FORMAT_LIST].replace(".", " ").replace(",", " ")
        format_set = set(out_format_list.split())
    else:
        format_set = set()

    if cfg[BASENAME] and (cfg[BASENAME] != DEF_BASENAME):
        # If cfg[BASENAME] is not just the base name, make it so, saving a dir or ext in their spots
        out_path, base_name = os.path.split(cfg[BASENAME])
        if out_path and cfg[OUT_DIR]:
            cfg[OUT_DIR] = os.path.join(cfg[OUT_DIR], out_path)
        elif out_path:
            cfg[OUT_DIR] = out_path
        base, ext = os.path.splitext(base_name)
        cfg[BASENAME] = base
        format_set.add(ext.replace(".", ""))

    if len(format_set) > 0:
        for format_type in format_set:
            if format_type in OUT_TYPE_LIST:
                cfg[SAVE_FILES] = True
                cfg[format_type] = True
            else:
                raise InvalidDataError(f"Invalid extension provided: '{format_type}'. The currently supported types "
                                       f"are: '{OUT_TYPE_STR}'")

    # if out_dir does not already exist, recreate it, only if we will actually need it
    if cfg[SAVE_FILES] and cfg[OUT_DIR]:
        make_dir(cfg[OUT_DIR])


def main(argv=None):
    """
    Runs the main program.

    :param argv: The command line arguments.
    :return: The return code for the program's termination.
    """
    print(OPENING_MSG)
    args, ret = parse_cmdline(argv)
    if ret != GOOD_RET or args is None:
        return ret

    cfg = args.config

    try:
        # tests at the beginning to catch errors early
        validate_input(cfg)

        for add_rate in cfg[ADD_RATES]:
            sg_adjs = []
            add_rate_str = f'{add_rate:.{3}g}'.replace("+", "").replace(".", "-")
            # todo: remove MANUSCRIPT_RATES when debugging done
            if cfg[RXN_RATES] == MANUSCRIPT_RATES:
                add_rate_str += "_m"
            for sg_ratio in cfg[SG_RATIOS]:
                # the initialized lists below are for storing repeats
                bond_types = defaultdict(list)
                num_monos = []
                num_oligs = []
                adj_repeats = []

                for rep in range(cfg[NUM_REPEATS]):

                    # decide on initial monomers, based on given sg_ratio, and create initial oxidation events
                    initial_events, initial_state = initiate_state(add_rate, cfg, rep, sg_ratio)

                    # begin simulation
                    result = run_kmc(cfg[RXN_RATES], initial_state, initial_events, n_max=cfg[MAX_MONOS],
                                     t_max=cfg[SIM_TIME], sg_ratio=sg_ratio, dynamics=cfg[DYNAMICS])

                    if cfg[DYNAMICS]:
                        last_adj = result[ADJ_MATRIX][-1]
                        last_mono_list = result[MONO_LIST][-1]
                        (bond_type_dict, olig_monos_dict, sum_monos_list, olig_count_dict,
                         sum_count_list) = get_bond_type_v_time_dict(result[ADJ_MATRIX], sum_len_larger_than=2)

                        for bond_type in BOND_TYPE_LIST:
                            bond_types[bond_type].append(bond_type_dict[bond_type])
                        num_monos.append(olig_count_dict[1])
                        num_oligs.append(sum_count_list)

                    else:
                        last_adj = result[ADJ_MATRIX]
                        last_mono_list = result[MONO_LIST]

                    adj_repeats.append(last_adj)

                    # show results
                    summary = analyze_adj_matrix(last_adj)
                    adj_analysis_to_stdout(summary)

                    # Outputs
                    produce_output(last_adj, last_mono_list, cfg)

                    # save for potential plotting
                sg_adjs.append(adj_repeats)

                # Now that all repeats done, create plots for dynamics, if applicable
                if cfg[DYNAMICS]:
                    # create plots of num mon & olig vs timestep, and % bond time v timestep
                    create_dynamics_plots(add_rate_str, bond_types, cfg, num_monos, num_oligs, sg_ratio)
            if cfg[PLOT_BONDS]:
                create_bond_v_sg_plots(add_rate_str, cfg, sg_adjs)

    except (InvalidDataError, KeyError) as e:
        warning(e)
        return INVALID_DATA

    return GOOD_RET  # success


if __name__ == '__main__':
    status = main()
    sys.exit(status)



# !/usr/bin/env python
# coding=utf-8

"""
Code base for simulating the in planta polymerization of monolignols through Gillepsie algorithm adaptations.
Added the visualization tools here.
"""

import re
import copy
import numpy as np
from collections import (defaultdict, OrderedDict)
from scipy import triu
from scipy.sparse import dok_matrix
from rdkit.Chem.Draw.MolDrawing import DrawingOptions
from common_wrangler.common import (InvalidDataError, create_out_fname, warning, round_sig_figs)
from ligninkmc.kmc_common import (Event, Monomer, AO4, B1, B5, BB, BO4, C5C5, C5O4, OX, Q, GROW, TIME, OLIGOMER,
                                  MONOMER, AFFECTED, ADJ_MATRIX, MONO_LIST, MAX_NUM_DECIMAL, ATOMS, BONDS,
                                  G, S, H, C, S4, G4, G7, B1_ALT, CHAIN_LEN, CHAIN_MONOS, CHAIN_BRANCHES,
                                  CHAIN_BRANCH_COEFF, RCF_BONDS, RCF_YIELDS, RCF_MONOS, RCF_BRANCHES, RCF_BRANCH_COEFF,
                                  DEF_TCL_FNAME, DEF_CHAIN_ID, DEF_PSF_FNAME, DEF_TOPPAR, INT_TO_TYPE_DICT)

DrawingOptions.bondLineWidth = 1.2
S7 = 'S7'


def find_fragments(adj):
    """
    Implementation of a modified depth first search on the adjacency matrix provided to identify isolated graphs within
    the superstructure. This allows us to easily track the number of isolated fragments and the size of each of these
    fragments. This implementation does not care about the specific values within the adjacency matrix, but effectively
    treats the adjacency matrix as boolean.

    :param adj: dok_matrix  -- NxN sparse matrix in dictionary of keys format that contains all of the connectivity
        information for the current lignification state
    :return: two lists where the list indices of each correspond to a unique fragment:
                A list of sets: the list contains a set for each fragment, comprised of the unique integer identifiers
                                for the monomers contained within the fragment,
                A list of ints containing the number of number of branch points found in each fragment
    """
    remaining_nodes = list(range(adj.get_shape()[0]))
    current_node = 0
    connected_fragments = [set()]
    connection_stack = []

    branches_in_frags = []
    num_branches = 0

    csr_adj = adj.tocsr(copy=True)

    while current_node is not None:
        # Indicate that we are currently visiting this node by removing it
        remaining_nodes.remove(current_node)

        # Add to the current_fragment
        current_fragment = connected_fragments[-1]

        # Look for what's connected to this row
        connections = {node for node in csr_adj[current_node].indices}
        # if more than two units are connected, there is a branch
        len_connections = len(connections)
        if len_connections > 2:
            num_branches += len_connections - 2

        # Add these connections to our current fragment
        current_fragment.update({current_node})

        # Visit any nodes that the current node is connected to that still need to be visited
        connection_stack.extend([node for node in connections if (node in remaining_nodes and
                                                                  node not in connection_stack)])

        # Get the next node that should be visited
        if len(connection_stack) != 0:
            current_node = connection_stack.pop()
        elif len(remaining_nodes) != 0:
            current_node = remaining_nodes[0]
            # great ready for next fragment
            connected_fragments.append(set())
            branches_in_frags.append(num_branches)
            num_branches = 0
        else:
            current_node = None
            branches_in_frags.append(num_branches)
    return connected_fragments, branches_in_frags


def fragment_size(frags):
    """
    A rigorous way to analyze_adj_matrix the size of fragments that have been identified using the find_fragments(adj)
    tool. Makes a dictionary of monomer identities mapped to the length of the fragment that contains them.

    Example usage:
    > frags = [{0}, {1}]
    > result = fragment_size(frags)
    {0: 1, 1: 1}

    > frags = [{0, 4, 2}, {1, 3}]
    > result = fragment_size(frags)
    {0: 3, 2: 3, 4: 3, 1: 2, 3: 2}

    > frags = [{0, 1, 2, 3, 4}]
    > result = fragment_size(frags)
    {0: 5, 1: 5, 2: 5, 3: 5, 4: 5}

    :param frags: list of sets; the set (list) of monomer identifier sets that were output from
                  find_fragments, or the monomers that are connected to each other
    :return: dict mapping the integer identity of each monomer to the length of the fragment that it is found in
    """
    # Note: this is only used by the test, so maybe delete
    sizes = {}
    for fragment in frags:
        length = len(fragment)
        for node in fragment:
            sizes[node] = length
    return sizes


def quick_frag_size(monomer):
    """
    An easy check on a specific monomer to tell if it is a monomer or involved in an oligomer. This is used over the
    detailed fragment_size(frags) calculation in the simulation of lignification for performance benefits.

    :param monomer: Monomer object that we want to know if it is bound to anything else
                     (i.e. if it is truly a monomer still)
    :return: string, either 'monomer' or 'oligomer' (as the global variable) if it is connected to nothing else,
             or isn't respectively
    """
    if monomer.type == G and monomer.open == {4, 5, 8}:  # Guaiacol monomer
        return MONOMER
    elif monomer.type == S and monomer.open == {4, 8}:  # Syringol monomer
        return MONOMER
    elif monomer.type == C and monomer.open == {4, 5, 8}:  # Caffeoyl monomer
        return MONOMER
    return OLIGOMER


def break_bond_type(adj, bond_type):
    """
    Function for removing all of a certain type of bond from the adjacency matrix. This is primarily used for the
    analysis at the end of the simulations when in silico RCF should occur. The update happens via conditional removal
    of the matching values in the adjacency matrix.

    Example use cases:
    > a = dok_matrix((5,5))
    > a[1,0] = 4; a[0,1] = 8; a[2,3] = 8; a[3,2] = 8;
    > break_bond_type(a, BO4).todense()
    [[0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0],
     [0, 0, 0, 8, 0],
     [0, 0, 8, 0, 0],
     [0, 0, 0, 0, 0]]

    > a = dok_matrix([[0, 4, 0, 0, 0],
    >                 [8, 0, 1, 0, 0],
    >                 [0, 8, 0, 0, 0],
    >                 [0, 0, 0, 0, 0],
    >                 [0, 0, 0, 0, 0]])
    > break_bond_type(a, B1_ALT).todense()
    [[0, 0, 0, 0, 0],
     [0, 0, 1, 0, 0],
     [0, 8, 0, 0, 0],
     [0, 0, 0, 0, 0],
     [0, 0, 0, 0, 0]]

    :param adj: dok_matrix, the adjacency matrix for the lignin polymer that has been simulated, and needs
        certain bonds removed
    :param bond_type: str, the string containing the bond type that should be broken. These are the standard
        nomenclature, except for B1_ALT, which removes the previous bond between the beta position and another monomer
        on the monomer that is bound through 1
    :return: dok_matrix, new adjacency matrix after bonds were broken
    """
    # Copy the matrix into a new matrix
    new_adj = adj.todok(copy=True)

    breakage = {B1: (lambda row, col: (adj[(row, col)] == 1 and adj[(col, row)] == 8) or (adj[(row, col)] == 8 and
                                                                                          adj[(col, row)] == 1)),
                B1_ALT: (lambda row, col: (adj[(row, col)] == 1 and adj[(col, row)] == 8) or
                                          (adj[(row, col)] == 8 and adj[(col, row)] == 1)),
                B5: (lambda row, col: (adj[(row, col)] == 5 and adj[(col, row)] == 8) or (adj[(row, col)] == 8 and
                                                                                          adj[(col, row)] == 5)),
                BO4: (lambda row, col: (adj[(row, col)] == 4 and adj[(col, row)] == 8) or (adj[(row, col)] == 8 and
                                                                                           adj[(col, row)] == 4)),
                AO4: (lambda row, col: (adj[(row, col)] == 4 and adj[(col, row)] == 7) or (adj[(row, col)] == 7 and
                                                                                           adj[(col, row)] == 4)),
                C5O4: (lambda row, col: (adj[(row, col)] == 4 and adj[(col, row)] == 5) or (adj[(row, col)] == 5 and
                                                                                            adj[(col, row)] == 4)),
                BB: (lambda row, col: (adj[(row, col)] == 8 and adj[(col, row)] == 8)),
                C5C5: (lambda row, col: (adj[(row, col)] == 5 and adj[(col, row)] == 5))}

    for adj_bond_loc in adj.keys():
        adj_row = adj_bond_loc[0]
        adj_col = adj_bond_loc[1]

        if breakage[bond_type](adj_row, adj_col) and bond_type != B1_ALT:
            new_adj[(adj_row, adj_col)] = 0
            new_adj[(adj_col, adj_row)] = 0
        elif breakage[bond_type](adj_row, adj_col):
            if adj[(adj_row, adj_col)] == 1:
                # The other 8 is in this row
                remove_prev_bond(adj, adj_row, new_adj)
            else:
                # The other 8 is in the other row
                remove_prev_bond(adj, adj_col, new_adj)
    return new_adj


def remove_prev_bond(adj, search_loc, new_adj):
    idx = 0  # make IDE happy
    data = adj.tocoo().getrow(search_loc).data
    cols = adj.tocoo().getrow(search_loc).indices
    for i, idx in enumerate(cols):
        if data[i] == 8:
            break
    new_adj[(search_loc, idx)] = 0
    new_adj[(idx, search_loc)] = 0


def count_bonds(adj):
    """
    Counter for the different bonds that are present in the adjacency matrix. Primarily used for getting easy analysis
    of the properties of a simulated lignin from the resulting adjacency matrix.

    :param adj: dok_matrix   -- the adjacency matrix for the lignin polymer that has been simulated
    :return: OrderedDict mapping bond strings to the frequency of that specific bond
    """
    bound_count_dict = OrderedDict({BO4: 0,  BB: 0, B5: 0, B1: 0, C5O4: 0, AO4: 0, C5C5: 0})
    bonding_dict = {(4, 8): BO4, (8, 4): BO4, (8, 1): B1, (1, 8): B1, (8, 8): BB, (5, 5): C5C5,
                    (8, 5): B5, (5, 8): B5, (7, 4): AO4, (4, 7): AO4, (5, 4): C5O4, (4, 5): C5O4}

    adj_array = triu(adj.toarray(), k=1)

    # Don't double count by looking only at the upper triangular keys
    for el in dok_matrix(adj_array).keys():
        row = el[0]
        col = el[1]

        bond = (adj[(row, col)], adj[(col, row)])
        bound_count_dict[bonding_dict[bond]] += 1

    return bound_count_dict


def count_oligomer_yields(adj):
    """
    Use the depth first search implemented in find_fragments(adj) to locate individual fragments and branching
    Related values are also calculated.

    :param adj: scipy dok_matrix, the adjacency matrix for the lignin polymer that has been simulated
    :return: four dicts: an OrderedDict for olig_len_dict (olig_len: num_oligs); the keys are common to all
                             dicts so one ordered dict should be sufficient. The other three dicts are:
                                 olig_length: the total number of monomers involved in oligomers
                                 olig_length: total number of branch points in oligomers of that length
                                 olig_length: the branching coefficient for the oligomers of that length
    """
    oligomers, branches_in_oligs = find_fragments(adj)

    temp_olig_len_dict = defaultdict(int)
    temp_olig_branch_dict = defaultdict(int)
    for oligomer, num_branches in zip(oligomers, branches_in_oligs):
        temp_olig_len_dict[len(oligomer)] += 1
        temp_olig_branch_dict[len(oligomer)] += num_branches

    # create one ordered dict, and three regular dicts
    olig_lengths = list(temp_olig_len_dict.keys())
    olig_lengths.sort()
    olig_len_dict = OrderedDict()
    olig_monos_dict = {}
    olig_branch_dict = {}
    olig_branch_coeff_dict = {}
    for olig_len in olig_lengths:
        num_oligs = temp_olig_len_dict[olig_len]
        num_monos_in_olig_length = num_oligs * olig_len
        num_branches = temp_olig_branch_dict[olig_len]
        olig_len_dict[olig_len] = num_oligs
        olig_monos_dict[olig_len] = num_monos_in_olig_length
        olig_branch_dict[olig_len] = num_branches
        olig_branch_coeff_dict[olig_len] = num_branches / num_monos_in_olig_length

    return olig_len_dict, olig_monos_dict, olig_branch_dict, olig_branch_coeff_dict


def analyze_adj_matrix(adjacency):
    """
    Performs the analysis for a single simulation to extract the relevant macroscopic properties, such as both the
    simulated frequency of different oligomer sizes and the number of each different type of bond before and after in
    silico RCF. The specific code to handle each of these properties is written in the count_bonds(adj) and
    count_oligomer_yields(adj) specifically.

    :param adjacency: scipy dok_matrix  -- the adjacency matrix for the lignin polymer that has been simulated
    :return: A dictionary of results: Chain Lengths, RCF Yields, Bonds, and RCF Bonds
    """

    # Remove any excess b1 bonds from the matrix, e.g. bonds that should be
    # broken during synthesis
    adjacency = break_bond_type(adjacency, B1_ALT)

    # Examine the initial polymers before any bonds are broken
    olig_yield_dict, olig_monos_dict, olig_branch_dict, olig_branch_coeff_dict = count_oligomer_yields(adjacency)
    bond_distributions = count_bonds(adjacency)

    # Simulate the RCF process at complete conversion by breaking all of the
    # alkyl C-O bonds that were formed during the reaction
    rcf_adj = break_bond_type(break_bond_type(break_bond_type(adjacency, BO4), AO4), C5O4)

    # Now count the bonds and yields remaining
    rcf_yield_dict, rcf_monos_dict, rcf_branch_dict, rcf_branch_coeff_dict = count_oligomer_yields(rcf_adj)
    rcf_bonds = count_bonds(rcf_adj)

    return {BONDS: bond_distributions, CHAIN_LEN: olig_yield_dict, CHAIN_MONOS:  olig_monos_dict,
            CHAIN_BRANCHES: olig_branch_dict, CHAIN_BRANCH_COEFF: olig_branch_coeff_dict,
            RCF_BONDS: rcf_bonds, RCF_YIELDS: rcf_yield_dict, RCF_MONOS: rcf_monos_dict,
            RCF_BRANCHES: rcf_branch_dict, RCF_BRANCH_COEFF: rcf_branch_coeff_dict}


def update_events(state_dict, adj, last_event, event_dict, rate_vec, rate_dict, max_mon=500):
    """
    This method determines what the possible events are in a given state, where the state is the current simulation
    state. Most of the additional parameters in this method are added for performance benefits rather than necessity.

    :param state_dict: dict that maps the index of each monomer in the simulation to the monomer itself and the
        event_dict that would be effected by a change to the monomer key. This makes it easy to quickly
        determine which of the event_dict in the simulation need to be updated and which should not be changed.
    :param adj: dok_matrix  -- The current state of the simulation represented by the adjacency matrix containing all
        of the monomers and the bonds between them (if any)
    :param last_event: The previous Event that occurred, which will tell us what monomers were effected. When combined
        with the state dictionary, this allows for efficient updating of the set of event_dict that are possible
    :param event_dict: dict  -- The set of all possible unique event_dict that must be updated and returned from this
        method, implemented in a hash map where the event hash value is the key
    :param rate_vec: dict  -- The rates of all of the unique event_dict implemented in a hash map where the Event
        hash value is the key
    :param rate_dict: dict  -- The dictionary of the possible rates involved in each reaction
    :param max_mon: int -- The maximum number of monomers that should be stored in the simulation
    :return: n/a, updates state_dict, adj, event_dict, rate_vec
    """

    # Map the monomer active state to the possible event_dict it can do
    possible_events = {0: [[OX, 1, rate_dict[OX]]],
                       4: [[B1, 2, rate_dict[B1], [1, 8]],
                           [C5O4, 2, rate_dict[C5O4], [4, 5]],
                           [AO4, 2, rate_dict[AO4], [4, 7]],
                           [BO4, 2, rate_dict[BO4], [4, 8]],
                           [C5C5, 2, rate_dict[C5C5], [5, 5]],
                           [B5, 2, rate_dict[B5], [5, 8]],
                           [BB, 2, rate_dict[BB], [8, 8]]],
                       7: [[Q, 1, rate_dict[Q]],
                           [AO4, 2, rate_dict[AO4], [7, 4]]],
                       -1: [[]]
                       }
    # Only do these for bonding and oxidation event_dict, any growth does not actually change the possible event_dict
    if last_event.key != GROW:
        # Remove the last event that we just did from the set of event_dict that can be performed
        last_event_hash = hash(last_event)
        del (event_dict[last_event_hash])
        del (rate_vec[last_event_hash])

        # Make sure to keep track of which partners have been "cleaned" already - i.e. what monomers have already had
        #     all of the old event_dict removed
        cleaned_partners = set()

        # Get indices of monomers that were acted upon
        affected_monomers = last_event.index

        # Update event_dict from the perspective of each monomer that was just affected individually
        # This prevents having to re-search the entire space every time, saving significant computation
        for mon_id in affected_monomers:
            # Get the affected monomer
            mon = state_dict[mon_id][MONOMER]

            # Get the sets of activated monomers that we could bind with
            ox = set()
            quinone = set()

            other_ids = [x for x in state_dict if x != mon_id]
            for other in other_ids:
                other_mon_type = state_dict[other][MONOMER]
                # Don't allow connections that would cyclize the polymer!
                if other_mon_type.active == 4 and other_mon_type.identity not in mon.connectedTo:
                    ox.add(other_mon_type)
                elif other_mon_type.active == 7 and other_mon_type.identity not in mon.connectedTo:
                    quinone.add(state_dict[other][MONOMER])
            bonding_partners = {BO4: ox, B5: ox, C5O4: ox, C5C5: ox, BB: ox, B1: ox, AO4: quinone}

            # Obtain the event_dict that are affected by a change to the monomer that was just acted on
            events_to_be_modified = state_dict[mon_id][AFFECTED]

            # Get the codes for the event_dict that are possible based on how the current monomer behaves
            active_pos = mon.active
            new_event_list = possible_events[active_pos]

            # Take any event_dict to be modified out of the set so that we can replace with the updated event_dict
            for event in events_to_be_modified:
                ev_hash = hash(event)
                if ev_hash in event_dict:
                    del (event_dict[ev_hash])
                    del (rate_vec[ev_hash])

            # Overwrite the old event_dict that could have been modified from this monomer being updated
            state_dict[mon_id][AFFECTED] = set()
            cur_n, _ = adj.get_shape()

            for rxn_event in new_event_list:
                if rxn_event and rxn_event[1] == 1:  # Unimolecular reaction event
                    size = quick_frag_size(mon)

                    rate = rxn_event[2][mon.type][size] / cur_n

                    # Add the event to the event_dict modifiable by changing the monomer, and update the set of all
                    # event_dict at the next time step
                    state_dict[mon_id][AFFECTED].add(Event(rxn_event[0], [mon.identity], rate))

                elif rxn_event and rxn_event[1] == 2:  # Bimolecular reaction event
                    bond = tuple(rxn_event[3])
                    alt = (bond[1], bond[0])
                    for partner in bonding_partners[rxn_event[0]]:
                        # Sanitize the set of event_dict that can be effected
                        if partner not in cleaned_partners:
                            # Remove any old event_dict from
                            state_dict[partner.identity][AFFECTED].difference_update(events_to_be_modified)
                            cleaned_partners.add(partner)

                        index = [mon.identity, partner.identity]
                        back = [partner.identity, mon.identity]

                        # Add the bond from one active unit to the other in the default config
                        size = (quick_frag_size(mon), quick_frag_size(partner))
                        if bond[0] in mon.open and bond[1] in partner.open:
                            try:
                                # todo: delete when questions re olig-olig b04 bond is resolved
                                if size == (OLIGOMER, OLIGOMER):
                                    if rxn_event[0] == BO4:
                                        print(f"{rxn_event[0]} reaction between oligomers with {mon.identity} and "
                                              f"{partner.identity}")
                                rate = rxn_event[2][(mon.type, partner.type)][size] / (cur_n ** 2)
                            except KeyError:
                                raise InvalidDataError(f"Error while attempting to update event_dict: event "
                                                       f"{rxn_event[0]} between indices {mon.identity} and "
                                                       f"{partner.identity} ")

                            # Add this to both the monomer and it's bonding partners list of event_dict that need to be
                            # modified upon manipulation of either monomer
                            # this -> other
                            state_dict[mon_id][AFFECTED].add(Event(rxn_event[0], index, rate, bond))
                            state_dict[partner.identity][AFFECTED].add(Event(rxn_event[0], index, rate, bond))

                            # Switch the order
                            # other -> this
                            state_dict[mon_id][AFFECTED].add(Event(rxn_event[0], back, rate, alt))
                            state_dict[partner.identity][AFFECTED].add(Event(rxn_event[0], back, rate, alt))

                        # Add the bond from one monomer to the other in the reverse config if not symmetric
                        if rxn_event[0] != BB and rxn_event[0] != C5C5:  # non-symmetric bond
                            if bond[1] in mon.open and bond[0] in partner.open:
                                # Adjust the rate using the correct monomer types
                                try:
                                    rate = rxn_event[2][(partner.type, mon.type)][(size[1], size[0])] / (cur_n ** 2)
                                except KeyError:
                                    raise InvalidDataError(f"Error on determining the rate for rxn_event type "
                                                           f"{rxn_event[0]}, bonding index {mon.identity} to "
                                                           f"{partner.identity}")
                                # this -> other alt
                                state_dict[mon_id][AFFECTED].add(Event(rxn_event[0], index, rate, alt))
                                state_dict[partner.identity][AFFECTED].add(Event(rxn_event[0], index, rate, alt))

                                # Switch the order
                                # other -> this alt
                                state_dict[mon_id][AFFECTED].add(Event(rxn_event[0], back, rate, bond))
                                state_dict[partner.identity][AFFECTED].add(Event(rxn_event[0], back, rate, bond))
                    # END LOOP OVER PARTNERS
                # END UNIMOLECULAR/BIMOLECULAR BRANCH
            # END LOOP OVER NEW REACTION POSSIBILITIES
            for event in state_dict[mon_id][AFFECTED]:
                ev_hash = hash(event)
                event_dict[ev_hash] = event
                rate_vec[ev_hash] = event.rate
        # END LOOP OVER MONOMERS THAT WERE AFFECTED BY LAST EVENT
    else:
        cur_n, _ = adj.get_shape()

        # If the system has grown to the maximum size, delete the event for adding more monomers
        if cur_n >= max_mon:
            last_event_hash = hash(last_event)
            del (event_dict[last_event_hash])
            del (rate_vec[last_event_hash])

        # Reflect the larger system volume
        for i in rate_vec:
            if event_dict[i].key != GROW:
                rate_vec[i] = rate_vec[i] * (cur_n - 1) / cur_n

        # Add an event to oxidize the monomer that was just added to the simulation
        oxidation = Event(OX, [cur_n - 1], rate_dict[OX][state_dict[cur_n - 1][MONOMER].type][MONOMER])
        state_dict[cur_n - 1][AFFECTED].add(oxidation)
        ev_hash = hash(oxidation)
        event_dict[ev_hash] = oxidation
        rate_vec[ev_hash] = oxidation.rate / cur_n


def connect_monos(mon1, mon2):
    if mon1.parent == mon1:
        if mon2.parent == mon2:
            parent = mon1 if mon1.identity <= mon2.identity else mon2
            mon1.parent = parent
            mon2.parent = parent
            parent.size = mon1.size + mon2.size
            return parent
        else:
            parent = connect_monos(mon1, mon2.parent)
            mon1.parent = parent
            mon2.parent = parent
            return parent
    else:
        parent = connect_monos(mon1.parent, mon2)
        mon2.parent = parent
        mon1.parent = parent
        return parent


def do_event(event, state, adj, sg_ratio=None, random_seed=None):
    """
    The second key component of the lignin implementation of the Monte Carlo algorithm, this method actually executes
    the chosen event on the current state and modifies it to reflect the updates.

    :param event: The event object that should be executed on the current state
    :param state: dict, The dictionary of dictionaries that contains the state information for each monomer
    :param adj: dok_matrix, The adjacency matrix in the current state
    :param sg_ratio: float needed if and only if:
                         a) there are S and G and only S and G, and
                         b) new state_dict will be added
    :param random_seed: if a positive integer is provided, it will be used for reproducible results (for testing)
    :return: N/A - mutates the list of state and adjacency matrix instead
    """

    indices = event.index
    # state is indirectly updated by changing monomers
    monomers = [state[i][MONOMER] for i in state]
    if len(indices) == 2:  # Doing bimolecular reaction, need to adjust adj

        # Get the tuple of values corresponding to bond and state updates and
        # unpack them
        vals = event.eventDict[event.key]
        state_updates = vals[0]
        bond_updates = event.bond
        order = event.activeDict[bond_updates]

        # Get the monomers that were being reacted in the correct order
        mon0 = monomers[indices[0]]
        mon1 = monomers[indices[1]]

        connect_monos(mon0, mon1)

        # Make the update to the state and adjacency matrix,
        # Rows are perspective of bonds FROM indices[0] and columns perspective of bonds TO indices[0]
        adj[(indices[0], indices[1])] = bond_updates[0]
        adj[(indices[1], indices[0])] = bond_updates[1]

        # remove the position that was just active
        mon0.open -= {bond_updates[0]}
        mon1.open -= {bond_updates[1]}

        # Update the activated nature of the monomer
        mon0.active = state_updates[order[0]]
        mon1.active = state_updates[order[1]]

        # Add any additional opened positions based on what just reacted
        mon0.open |= set(vals[1 + order[0]])
        mon1.open |= set(vals[1 + order[1]])

        if mon0.active == 7 and mon1.type == C:
            mon0.active = 0
            mon0.open -= {7}

        if mon1.active == 7 and mon0.type == C:
            mon1.active = 0
            mon1.open -= {7}

        # Decided to break bond between alpha and ring position later (i.e. after all synthesis occurred) when a B1
        # bond is formed This is primarily to make it easier to see what the fragment that needs to break is for
        # visualization purposes
        mon0.connectedTo.update(mon1.connectedTo)
        for mon in monomers:
            if mon.identity in mon0.connectedTo:
                mon.connectedTo = mon0.connectedTo

    elif len(indices) == 1:
        if event.key == Q:
            mon = monomers[indices[0]]
            mon.active = 0
            mon.open.remove(7)
            mon.open.add(1)
        elif event.key == OX:
            mon = monomers[indices[0]]

            # Make the monomer appear oxidized
            mon.active = 4
        else:
            raise InvalidDataError(f'Unexpected event: {event.key} for index {indices[0]}')
    else:
        if event.key == GROW:
            current_size, _ = adj.get_shape()

            # Add another monomer to the adjacency matrix
            adj.resize((current_size + 1, current_size + 1))

            # Add another monomer to the state
            if monomers and monomers[-1].type == C:
                mon_type = C
            else:
                try:
                    pct = sg_ratio / (1 + sg_ratio)
                    if random_seed:
                        # to prevent the same choice every iteration, add a changing integer
                        np.random.seed(random_seed + current_size)
                        rand_num = np.around(np.random.rand(), MAX_NUM_DECIMAL)
                    else:
                        rand_num = np.random.rand()
                    mon_type = INT_TO_TYPE_DICT[int(rand_num < pct)]
                except TypeError:
                    if sg_ratio is None:
                        sg_note = " the default value 'None'."
                    else:
                        sg_note = f": {sg_ratio}"
                    raise InvalidDataError(f"A numeric sg_ratio must be supplied. Instead, found{sg_note}")

            new_mon = Monomer(mon_type, current_size)
            state[current_size] = {MONOMER: new_mon, AFFECTED: set()}


def run_kmc(rate_dict, initial_state, initial_events, n_max=10, t_max=10, dynamics=False, random_seed=None,
            sg_ratio=None):
    """
    Performs the Gillespie algorithm using the specific event and update implementations described by do_event and
    update_events specifically. The initial state and event_dict in that state are constructed and passed to the run_kmc
    method, along with the possible rates of different bond formation event_dict, the maximum number of monomers that
    should be included in the simulation and the total simulation time.

    :param rate_dict:  dict   -- contains the reaction rate of each of the possible event_dict
    :param initial_state: dict  -- The dictionary mapping the index of each monomer to a dictionary with the monomer
        and the set of event_dict that a change to this monomer would impact
    :param initial_events: list of dicts dictionary -- The dictionary mapping event hash values to those event_dict
    :param n_max:   int   -- The maximum number of monomers in the simulation
    :param t_max: float -- The final simulation time (units depend on units of rates)
    :param dynamics: boolean -- if True, will keep values for every time step
    :param random_seed: None or hashable value to aid testing
    :param sg_ratio: needed if there is S and G and nothing else
    :return: dict with the simulation times, adjacency matrix, and list of monomers at the end of the simulation
    """
    # probably not going to use the initial_state again, but copying anyway so it remains pristine
    cur_state = copy.deepcopy(initial_state)

    # Current number of monomers
    num_monos = len(cur_state.keys())
    adj = dok_matrix((num_monos, num_monos))
    t = [0, ]

    # Calculate the rates of all of the event_dict available at the current state; ordered dict for consistency
    r_vec = OrderedDict()

    # Build the dictionary of event_dict
    event_dict = OrderedDict()
    for event in initial_events:
        event_hash = hash(event)
        r_vec[event_hash] = event.rate / num_monos
        event_dict[event_hash] = event

    if dynamics:
        adj_list = [adj.copy()]
        mon_list = [[copy.copy(cur_state[i][MONOMER]) for i in cur_state]]
    else:  # just to make IDE happy that won't use before defined
        adj_list = []
        mon_list = []

    # Run the Gillespie algorithm
    while t[-1] < t_max and len(event_dict) > 0:
        # Find the total rate for all of the possible event_dict and choose which event to do
        # Less direct conversion of r_vec to lists (keys in hashes, vals in all_rates) for consistency
        hashes = list(r_vec.keys())
        hashes.sort()
        all_rates = []
        for r_vec_key in hashes:
            all_rates.append(r_vec[r_vec_key])
        r_tot = np.sum(all_rates)

        if random_seed:
            # don't want same dt for every iteration, so add to seed with each iteration
            np.random.seed(random_seed + len(t))
            # don't let machine precision change the dt on different platforms
            rand_num = np.around(np.random.rand(), MAX_NUM_DECIMAL)
        else:
            rand_num = np.random.rand()
        j = np.random.choice(hashes, p=all_rates / r_tot)
        event = event_dict[j]
        # See how much time has passed before this event happened; rounding to reduce platform dependency
        dt = round_sig_figs((1 / r_tot) * np.log(1 / rand_num))
        t.append(t[-1] + dt)

        # Do the event and update the state
        do_event(event, cur_state, adj, sg_ratio, random_seed=random_seed)

        if dynamics:
            adj_list.append(adj.copy())
            mon_list.append([copy.copy(cur_state[i][MONOMER]) for i in cur_state])

        # Check the new state for what events are possible
        update_events(cur_state, adj, event, event_dict, r_vec, rate_dict, max_mon=n_max)

    if dynamics:
        return {TIME: t, MONO_LIST: mon_list, ADJ_MATRIX: adj_list}

    return {TIME: t, MONO_LIST: [cur_state[i][MONOMER] for i in cur_state], ADJ_MATRIX: adj}


# noinspection PyTypeChecker
def generate_mol(adj, node_list):
    """
    Based on standard molfile format https://www.daylight.com/meetings/mug05/Kappler/ctfile.pdf
    :param adj: dok_matrix
    :param node_list: list
    :return: mol_str, str in standard molfile
    """
    # define dictionary for atoms within each monomer
    atom_blocks = {G: ('C 0 0 0 0 \n' +  # 1
                       'C 0 0 0 0 \n' +  # 2
                       'C 0 0 0 0 \n' +  # 3
                       'C 0 0 0 0 \n' +  # 4
                       'C 0 0 0 0 \n' +  # 5
                       'C 0 0 0 0 \n' +  # 6
                       'C 0 0 0 0 \n' +  # 7
                       'C 0 0 0 0 \n' +  # 8
                       'C 0 0 0 0 \n' +  # 9
                       'O 0 0 0 0 \n' +  # 9-OH
                       'O 0 0 0 0 \n' +  # 3-OMe
                       'C 0 0 0 0 \n' +  # 3-OMe
                       'O 0 0 0 0 \n'),  # 4-OH
                   S: ('C 0 0 0 0 \n' +  # 1
                       'C 0 0 0 0 \n' +  # 2
                       'C 0 0 0 0 \n' +  # 3
                       'C 0 0 0 0 \n' +  # 4
                       'C 0 0 0 0 \n' +  # 5
                       'C 0 0 0 0 \n' +  # 6
                       'C 0 0 0 0 \n' +  # 7
                       'C 0 0 0 0 \n' +  # 8
                       'C 0 0 0 0 \n' +  # 9
                       'O 0 0 0 0 \n' +  # 9-OH
                       'O 0 0 0 0 \n' +  # 3-OMe
                       'C 0 0 0 0 \n' +  # 3-OMe
                       'O 0 0 0 0 \n' +  # 4-OH
                       'O 0 0 0 0 \n' +  # 5-OMe
                       'C 0 0 0 0 \n'),  # 5-OMe
                   C: ('C 0 0 0 0 \n' +  # 1
                       'C 0 0 0 0 \n' +  # 2
                       'C 0 0 0 0 \n' +  # 3
                       'C 0 0 0 0 \n' +  # 4
                       'C 0 0 0 0 \n' +  # 5
                       'C 0 0 0 0 \n' +  # 6
                       'C 0 0 0 0 \n' +  # 7
                       'C 0 0 0 0 \n' +  # 8
                       'C 0 0 0 0 \n' +  # 9
                       'O 0 0 0 0 \n' +  # 9-OH
                       'O 0 0 0 0 \n' +  # 3-OH
                       'O 0 0 0 0 \n'),  # 4-OH
                   G4: ('C 0 0 0 0 \n' +  # 1
                        'C 0 0 0 0 \n' +  # 2
                        'C 0 0 0 0 \n' +  # 3
                        'C 0 0 0 0 \n' +  # 4
                        'C 0 0 0 0 \n' +  # 5
                        'C 0 0 0 0 \n' +  # 6
                        'C 0 0 0 0 \n' +  # 7
                        'C 0 0 0 0 \n' +  # 8
                        'C 0 0 0 0 \n' +  # 9
                        'O 0 0 0 0 \n' +  # 9-OH
                        'O 0 0 0 0 \n' +  # 3-OMe
                        'C 0 0 0 0 \n' +  # 3-OMe
                        'O 0 0 0 0 RAD=2\n'),  # 4-O
                   S4: ('C 0 0 0 0 \n' +  # 1
                        'C 0 0 0 0 \n' +  # 2
                        'C 0 0 0 0 \n' +  # 3
                        'C 0 0 0 0 \n' +  # 4
                        'C 0 0 0 0 \n' +  # 5
                        'C 0 0 0 0 \n' +  # 6
                        'C 0 0 0 0 \n' +  # 7
                        'C 0 0 0 0 \n' +  # 8
                        'C 0 0 0 0 \n' +  # 9
                        'O 0 0 0 0 \n' +  # 9-OH
                        'O 0 0 0 0 \n' +  # 3-OMe
                        'C 0 0 0 0 \n' +  # 3-OMe
                        'O 0 0 0 0 RAD=2\n' +  # 4-O
                        'O 0 0 0 0 \n' +  # 5-OMe
                        'C 0 0 0 0 \n')}  # 5-OMe

    # Similarly define dictionary for bonds within each monomer -
    # NOTE: THESE MAY NEED TO CHANGE DEPENDING ON INTER-UNIT LINKAGES

    bond_blocks = {G7: ('1 1  2  \n' +  # Aromatic ring 1->2
                        '2 2  3  \n' +  # Aromatic ring 2->3
                        '1 3  4  \n' +  # Aromatic ring 3->4
                        '1 4  5  \n' +  # Aromatic ring 4->5
                        '2 5  6  \n' +  # Aromatic ring 5->6
                        '1 6  1  \n' +  # Aromatic ring 6->1
                        '2 1  7  \n' +  # Quinone methide propyl tail 1->A
                        '1 7  8  \n' +  # Propyl tail A->B
                        '1 8  9  \n' +  # Propyl tail B->G
                        '1 9  10 \n' +  # Gamma hydroxyl G->OH
                        '1 3  11 \n' +  # 3 methoxy 3->O
                        '1 11 12 \n' +  # 3 methoxy O->12
                        '2 4  13 \n'),  # 4 ketone 4->O
                   G: ('2 1  2  \n' +  # Aromatic ring 1->2
                       '1 2  3  \n' +  # Aromatic ring 2->3
                       '2 3  4  \n' +  # Aromatic ring 3->4
                       '1 4  5  \n' +  # Aromatic ring 4->5
                       '2 5  6  \n' +  # Aromatic ring 5->6
                       '1 6  1  \n' +  # Aromatic ring 6->1
                       '1 1  7  \n' +  # Ring - propyl tail 1->A
                       '2 7  8  \n' +  # Alkene propyl tail A->B
                       '1 8  9  \n' +  # Propyl tail B->G
                       '1 9  10 \n' +  # Gamma hydroxyl G->OH
                       '1 3  11 \n' +  # 3 methoxy 3->O
                       '1 11 12 \n' +  # 3 methoxy O->12
                       '1 4  13 \n'),  # 4 hydroxyl 4->OH
                   S7: ('1 1  2  \n' +  # Aromatic ring 1->2
                        '2 2  3  \n' +  # Aromatic ring 2->3
                        '1 3  4  \n' +  # Aromatic ring 3->4
                        '1 4  5  \n' +  # Aromatic ring 4->5
                        '2 5  6  \n' +  # Aromatic ring 5->6
                        '1 6  1  \n' +  # Aromatic ring 6->1
                        '2 1  7  \n' +  # Quinone methide 1->A
                        '1 7  8  \n' +  # Propyl tail A->B
                        '1 8  9  \n' +  # Propyl tail B->G
                        '1 9  10 \n' +  # Gamma hydroxyl G->OH
                        '1 3  11 \n' +  # 3 methoxy 3->O
                        '1 11 12 \n' +  # 3 methoxy O->12
                        '2 4  13 \n' +  # 4 ketone 4->O
                        '1 5  14 \n' +  # 5 methoxy 5->O
                        '1 14 15 \n'),  # 5 methoxy O->15
                   S: ('2 1  2  \n' +  # Aromatic ring 1->2
                       '1 2  3  \n' +  # Aromatic ring 2->3
                       '2 3  4  \n' +  # Aromatic ring 3->4
                       '1 4  5  \n' +  # Aromatic ring 4->5
                       '2 5  6  \n' +  # Aromatic ring 5->6
                       '1 6  1  \n' +  # Aromatic ring 6->1
                       '1 1  7  \n' +  # Ring - propyl tail 1->A
                       '2 7  8  \n' +  # Alkene propyl tail A->B
                       '1 8  9  \n' +  # Propyl tail B->G
                       '1 9  10 \n' +  # Gamma hydroxyl G->OH
                       '1 3  11 \n' +  # 3 methoxy 3->O
                       '1 11 12 \n' +  # 3 methoxy O->12
                       '1 4  13 \n' +  # 4 hydroxyl 4->OH
                       '1 5  14 \n' +  # 5 methoxy 5->O
                       '1 14 15 \n'),  # 5 methoxy O->15
                   C: ('2 1  2  \n' +  # Aromatic ring 1->2
                       '1 2  3  \n' +  # Aromatic ring 2->3
                       '2 3  4  \n' +  # Aromatic ring 3->4
                       '1 4  5  \n' +  # Aromatic ring 4->5
                       '2 5  6  \n' +  # Aromatic ring 5->6
                       '1 6  1  \n' +  # Aromatic ring 6->1
                       '1 1  7  \n' +  # Ring - propyl tail 1->A
                       '2 7  8  \n' +  # Alkene propyl tail A->B
                       '1 8  9  \n' +  # Propyl tail B->G
                       '1 9  10 \n' +  # Gamma hydroxyl G->OH
                       '1 3  11 \n' +  # 3 hydroxyl 3->O
                       '1 4  12 \n')}  # 4 hydroxyl 4->OH

    mol_str = '\n\n\n  0  0  0  0  0  0  0  0  0  0999 V3000\nM  V30 BEGIN CTAB\n'  # Header information
    mol_atom_blocks = 'M  V30 BEGIN ATOM\n'
    mol_bond_blocks = 'M  V30 BEGIN BOND\n'
    atom_line_num = 1
    bond_line_num = 1
    mono_start_idx_bond = []
    mono_start_idx_atom = []
    removed = {BONDS: 0, ATOMS: 0}

    site_positions = {1: {x: 0 for x in [G, S, C]},
                      4: {C: 11, S: 12, G: 12},
                      5: {x: 4 for x in [G, S, C]},
                      7: {x: 6 for x in [G, S, C]},
                      8: {x: 7 for x in [G, S, C]},
                      10: {x: 9 for x in [G, S, C]}}
    alpha_beta_alkene_location = 7
    alpha_ring_location = 6
    alpha = 7

    # to make IDE happy:
    atom_block = None
    bond_block = None

    # Build the individual monomers before they are linked by anything
    for i, mon in enumerate(node_list):
        if mon.type == G or mon.type == S:
            if mon.active == 0 or mon.active == -1:
                if mon.type == G:
                    atom_block = atom_blocks[G]
                    bond_block = bond_blocks[G]
                else:
                    atom_block = atom_blocks[S]
                    bond_block = bond_blocks[S]
            elif mon.active == 4:
                if mon.type == G:
                    atom_block = atom_blocks[G4]
                    bond_block = bond_blocks[G]
                else:
                    atom_block = atom_blocks[S4]
                    bond_block = bond_blocks[S]
            elif mon.active == 7:
                if mon.type == G:
                    atom_block = atom_blocks[G]
                    bond_block = bond_blocks[G7]
                else:
                    atom_block = atom_blocks[S]
                    bond_block = bond_blocks[S7]
        elif mon.type == C:
            atom_block = atom_blocks[C]
            bond_block = bond_blocks[C]
        else:
            raise ValueError("Expected monomer types are {LIGNIN_SUBUNITS} but encountered type '{mon.type}'")

        # Extract each of the individual atoms from this monomer to add to the aggregate file
        lines = atom_block.splitlines(keepends=True)

        # Figure out what atom and bond number this monomer is starting at
        mono_start_idx_bond.append(bond_line_num)
        mono_start_idx_atom.append(atom_line_num)

        # Loop through the lines of the atom block and add the necessary prefixes to the lines, using a continuing
        #     atom index
        for line in lines:
            mol_atom_blocks += f'M  V30 {atom_line_num} {line}'
            atom_line_num += 1
        # END ATOM AGGREGATION

        # Extract each of the individual bonds that defines the monomer skeleton and add it into the aggregate string
        lines = bond_block.splitlines(keepends=True)

        # Recall where this monomer started
        # So that we can add the defined bond indices to this start index to get the bond defs
        start_index = mono_start_idx_atom[-1] - 1

        # Loop through the lines of the bond block and add necessary prefixes to the lines, then modify as needed after
        for line in lines:
            # Extract the defining information about the monomer skeleton
            bond_vals = re.split(' +', line)

            # The first element is the bond order, followed by the indices of the atoms that are connected by this bond.
            # These indices need to be updated based on the true index of this monomer, not the 1 -> ~15 indices that
            #     it started with
            bond_order = bond_vals[0]
            bond_connects = [int(bond_vals[1]) + start_index, int(bond_vals[2]) + start_index]

            # Now save the true string for defining this bond, along with the cumulative index of the bond
            mol_bond_blocks += f'M  V30 {bond_line_num} {bond_order} {bond_connects[0]} {bond_connects[1]} \n'
            bond_line_num += 1
        # END BOND AGGREGATION
    # END MONOMER AGGREGATION

    # Now that we have all of the monomers in one string, we just need to add the bonds and atoms that are defined by
    #     the adjacency matrix
    bonds = mol_bond_blocks.splitlines(keepends=True)
    atoms = mol_atom_blocks.splitlines(keepends=True)

    break_alkene = {(4, 8): True, (5, 8): True, (8, 8): True, (5, 5): False, (1, 8): True, (4, 7): False, (4, 5): False}
    hydrate = {(4, 8): True, (5, 8): False, (8, 8): False, (5, 5): False, (1, 8): True, (4, 7): False, (4, 5): False}
    beta = {(4, 8): 1, (8, 4): 0, (5, 8): 1, (8, 5): 0, (1, 8): 1, (8, 1): 0}
    make_alpha_ring = {(4, 8): False, (5, 8): True, (8, 8): True, (5, 5): False, (1, 8): False, (4, 7): False,
                       (4, 5): False}

    # Start by looping through the adjacency matrix, one bond at a time (corresponds to a pair iterator)
    paired_adj = zip(*[iter(dict(adj))] * 2)

    for pair in paired_adj:
        # Find the types of bonds and indices associated with each of the elements in the adjacency matrix
        # Indices are extracted as tuples (row,col) and we just want the row for each
        mono_indices = [x[0] for x in pair]

        # Get the monomers corresponding to the indices in this bond
        mons = [None, None]

        for mon in node_list:
            if mon.identity == mono_indices[0]:
                mons[0] = mon
            elif mon.identity == mono_indices[1]:
                mons[1] = mon

        # Now just extract the bond types (the value from the adj matrix)
        bond_loc = [int(adj[p]) for p in pair]

        # Get the indices of the atoms being bound -> Count from where the monomer starts, and add however many is
        #     needed to reach the desired position for that specific monomer type and bonding site
        atom_indices = [mono_start_idx_atom[mono_indices[i]] + site_positions[bond_loc[i]][mons[i].type]
                        for i in range(2)]

        # Make the string to add to the molfile
        bond_string = f'M  V30 {bond_line_num} 1 {atom_indices[0]} {atom_indices[1]} \n'
        bond_line_num += 1

        # Append the newly created bond to the file
        bonds.append(bond_string)

        # Check if the alkene needs to be modified to a single bond
        bond_loc_tuple = tuple(sorted(bond_loc))
        if break_alkene[bond_loc_tuple]:
            for i in range(2):
                if adj[pair[i]] == 8 and mons[i].active != 7:  # Monomer index is bound through beta
                    # Find the bond index corresponding to alkene bond
                    alkene_bond_index = mono_start_idx_bond[mono_indices[i]] + alpha_beta_alkene_location - \
                                        removed[BONDS]

                    # Get all of the required bond information (index,order,monIdx1,monIdx2)
                    bond_vals = re.split(' +', bonds[alkene_bond_index])[2:]
                    try:
                        assert (int(bond_vals[0]) == alkene_bond_index + removed[BONDS])
                    except AssertionError:
                        print(f'Expected index: {bond_vals[0]}, Index obtained: {alkene_bond_index}')

                    # Decrease the bond order by 1
                    bonds[alkene_bond_index] = f'M  V30 {bond_vals[0]} 1 {bond_vals[2]} {bond_vals[3]} \n'

        # Check if we need to add water to the alpha position
        if hydrate[bond_loc_tuple] and 7 not in adj[mono_indices[beta[tuple(bond_loc)]]].values() and \
                mons[beta[tuple(bond_loc)]].active != 7:
            if mons[int(not beta[tuple(bond_loc)])].type != C:
                # We should actually only be hydrating BO4 bonds when the alpha position is unoccupied (handled by
                # second clause above)

                # Find the location of the alpha position
                alpha_idx = mono_start_idx_atom[mono_indices[beta[tuple(bond_loc)]]] + site_positions[7][
                    mons[beta[tuple(bond_loc)]].type]

                # Add the alpha hydroxyl O atom
                atoms.append(f'M  V30 {atom_line_num} O 0 0 0 0 \n')
                atom_line_num += 1

                # Make the bond
                bonds.append(f'M  V30 {bond_line_num} 1 {alpha_idx} {atom_line_num - 1} \n')
                bond_line_num += 1
            else:
                # Make the benzodioxane linkage
                alpha_idx = mono_start_idx_atom[mono_indices[beta[tuple(bond_loc)]]] + site_positions[7][
                    mons[beta[tuple(bond_loc)]].type]
                hydroxy_index = mono_start_idx_atom[mono_indices[int(not beta[tuple(bond_loc)])]] + (
                        site_positions[4][mons[beta[tuple(bond_loc)]].type] - 1)  # subtract 1 to move from 4-OH to 3-OH

                # Make the bond
                bonds.append(f'M  V30 {bond_line_num} 1 {alpha_idx} {hydroxy_index} \n')
                bond_line_num += 1

        # Check if there will be a ring involving the alpha position
        if make_alpha_ring[bond_loc_tuple]:
            other_site = {(5, 8): 4, (8, 8): 10}
            for i in range(2):
                if adj[pair[i]] == 8:  # This index is bound through beta (will get alpha connection)
                    # Find the location of the alpha position and the position that cyclizes with alpha
                    alpha_idx = mono_start_idx_atom[mono_indices[i]] + site_positions[7][mons[i].type]
                    other_idx = mono_start_idx_atom[mono_indices[int(not i)]] + \
                        site_positions[other_site[bond_loc_tuple]][mons[int(not i)].type]

                    bonds.append(f'M  V30 {bond_line_num} 1 {alpha_idx} {other_idx} \n')
                    bond_line_num += 1

        # For the B1 bond:
        #     1) Disconnect the original 1 -> A bond that existed from the not beta monomer
        #     2) Convert the new primary alcohol to an aldehyde
        if sorted(bond_loc) == [1, 8]:
            # TODO: make sure all B1 bonds are correctly forms
            warning("There are problems in how this program currently builds molecules with B1 bonds. Carefully "
                    "check any output.")
            index_for_one = int(not beta[tuple(bond_loc)])
            # Convert the alpha alcohol on one's tail to an aldehyde
            alpha_idx = mono_start_idx_atom[mono_indices[index_for_one]
                                            ] + site_positions[alpha][mons[index_for_one].type]

            # Temporarily join the bonds so that we can find the string
            temp = ''.join(bonds)
            matches = re.findall(f'M {2}V30 [0-9]+ 1 {alpha_idx} [0-9]+', temp)

            # Find the bond connecting the alpha to the alcohol
            others = []
            for possibility in matches:
                bound_atoms = re.split(' +', possibility)[4:]
                others.extend([int(x) for x in bound_atoms if int(x) != alpha_idx])

            # TODO: Fix the section below--it does not work; for now, simply catch the error and provide a
            #   descriptive failure message (instead of a stack trace)
            # The oxygen atom should have the greatest index of the atoms bound to the alpha position because it
            #     was added last
            try:
                oxygen_atom_index = max(others)
                bonds = re.sub(f'1 {alpha_idx} {oxygen_atom_index}',
                               f'2 {alpha_idx} {oxygen_atom_index}', temp).splitlines(keepends=True)

                # Find where the index for the bond is and remove it from the array
                alpha_ring_bond_index = mono_start_idx_bond[mono_indices[index_for_one]
                                                            ] + alpha_ring_location - removed[BONDS]
                del (bonds[alpha_ring_bond_index])
                removed[BONDS] += 1
            except ValueError:
                raise InvalidDataError("This program cannot currently generate a molecule with this beta-1 bond. Sorry!"
                                       "\n    Try running again, as this bond type is only occasionally created.")

    mol_bond_blocks = ''.join(bonds)
    mol_atom_blocks = ''.join(atoms)

    mol_atom_blocks += 'M  V30 END ATOM \n'
    mol_bond_blocks += 'M  V30 END BOND \n'
    counts = f'M  V30 COUNTS {atom_line_num - 1 - removed[ATOMS]} {bond_line_num - 1 - removed[BONDS]} 0 0 0\n'
    mol_str += counts + mol_atom_blocks + mol_bond_blocks + 'M  V30 END CTAB\nM  END'

    return mol_str


def write_patch(open_file, patch_name, segname, resid1, resid2=None):
    """
    Simple script to consistently format patch output for tcl script
    :param open_file: {TextIOWrapper}
    :param patch_name: str
    :param segname: str
    :param resid1: int
    :param resid2: int
    :return: what to write to file
    """
    if resid2:
        open_file.write(f"patch {patch_name} {segname}:{resid1} {segname}:{resid2}\n")
    else:
        open_file.write(f"patch {patch_name} {segname}:{resid1}\n")


def gen_tcl(orig_adj, monomers, tcl_fname=DEF_TCL_FNAME, psf_fname=DEF_PSF_FNAME, chain_id=DEF_CHAIN_ID,
            toppar_dir=DEF_TOPPAR, out_dir=None):
    """
    This takes a computed adjacency matrix and monomer list and writes out a script to generate a psf file of the
    associated structure, suitable for feeding into the LigninBuilder plugin of VMD
    (https://github.com/jvermaas/LigninBuilder).

    :param orig_adj: dok_matrix, Adjacency matrix generated by the kinetic Monte Carlo process
    :param monomers: list of Monomer objects, Monomer list from the kinetic Monte Carlo process
    :param tcl_fname: str, desired output filename
    :param psf_fname: str, desired basename for psf and pdb files to be generated by VMD
    :param chain_id: str, desired `chainID` to be used in the resulting output segment name for the generated lignin
    :param toppar_dir: location where the topology files top_lignin.top and top_all36_cgenff.rtf are expected
    :param out_dir: directory where the .tcl file should be saved
    :return:
    """
    adj = orig_adj.copy()
    # In the unlikely event that chain_id starts with a space (won't happen if provided through create_lignin),
    #     get rid of it, as it would cause problems. Also make it uppercase, as that is what PDBs use.
    chain_id = chain_id.strip().upper()
    # in the unlikely event that chain_id is empty, which would also cause problems, give it the default, instead
    #     of just failing at the end, after all the work of generating the lignin. Also truncate a too-long chainID.
    if not chain_id or len(chain_id) > 1:
        if len(chain_id) > 1:
            chain_id = chain_id[0]
        else:
            chain_id = DEF_CHAIN_ID
        warning(f"ChainID's for PDBs should be one character. Will use: '{chain_id}' as the chainID.")
    residue_letter = {G: 'G', S: 'S', H: 'H', C: 'C'}
    f_out = create_out_fname(tcl_fname, base_dir=out_dir)
    # add a mac/linux dir separator if there isn't already a directory separator, and if there is to be a subdirectory
    #   (not None or "")
    if toppar_dir and len(toppar_dir) > 0:
        if toppar_dir[-1] != '/' and (toppar_dir[-1] != '\\'):
            toppar_dir += "/"
    else:
        toppar_dir = ""
    with open(f_out, "w") as f:
        f.write(f"package require psfgen\n"
                f"topology {toppar_dir}{'top_all36_cgenff.rtf'}\n"
                f"topology {toppar_dir}{'top_lignin.top'}\n"
                f"segment {chain_id} {{\n")
        for monomer in monomers:
            resid = monomer.identity + 1
            res_name = residue_letter[monomer.type]
            f.write(f"    residue {resid} {res_name}\n")
        f.write(f"}}\n")

        for row in (adj == 1).nonzero()[0]:
            col = (adj.getrow(row) == 8).nonzero()[1]
            if len(col):
                col = col[0]
                adj[(row, col)] *= -1
        for bond_matrix_tuple in adj.keys():
            if bond_matrix_tuple[0] > bond_matrix_tuple[1]:
                continue
            psf_patch_resid1 = bond_matrix_tuple[0] + 1
            psf_patch_resid2 = bond_matrix_tuple[1] + 1
            flipped_bond_matrix_tuple = (bond_matrix_tuple[1], bond_matrix_tuple[0])
            bond_loc1 = int(adj[bond_matrix_tuple])
            bond_loc2 = int(adj[flipped_bond_matrix_tuple])
            if bond_loc1 == 8 and bond_loc2 == 4:  # Beta-O-4 linkage
                write_patch(f, "BO4", chain_id, psf_patch_resid1, psf_patch_resid2)
            elif bond_loc1 == 4 and bond_loc2 == 8:  # Reverse beta-O-4 linkage.
                write_patch(f, "BO4", chain_id, psf_patch_resid2, psf_patch_resid1)
            elif bond_loc1 == 8 and bond_loc2 == 5 and monomers[bond_matrix_tuple[1]].type == G:  # B5G linkage
                write_patch(f, "B5G", chain_id, psf_patch_resid1, psf_patch_resid2)
            elif bond_loc1 == 5 and bond_loc2 == 8 and monomers[bond_matrix_tuple[0]].type == G:  # Reverse B5G linkage
                write_patch(f, "B5G", chain_id, psf_patch_resid2, psf_patch_resid1)
            elif bond_loc1 == 8 and bond_loc2 == 5 and monomers[bond_matrix_tuple[1]].type == C:  # B5C linkage
                write_patch(f, "B5C", chain_id, psf_patch_resid1, psf_patch_resid2)
            elif bond_loc1 == 5 and bond_loc2 == 8 and monomers[bond_matrix_tuple[0]].type == C:  # Reverse B5C linkage
                write_patch(f, "B5C", chain_id, psf_patch_resid2, psf_patch_resid1)
            elif bond_loc1 == 5 and bond_loc2 == 5:  # 55 linkage
                write_patch(f, "B5C", chain_id, psf_patch_resid2, psf_patch_resid1)
            elif bond_loc1 == 7 and bond_loc2 == 4:  # alpha-O-4 linkage
                write_patch(f, "AO4", chain_id, psf_patch_resid1, psf_patch_resid2)
            elif bond_loc1 == 4 and bond_loc2 == 7:  # Reverse alpha-O-4 linkage
                write_patch(f, "AO4", chain_id, psf_patch_resid2, psf_patch_resid1)
            elif bond_loc1 == 4 and bond_loc2 == 5:  # 4O5 linkage
                write_patch(f, "4O4", chain_id, psf_patch_resid1, psf_patch_resid2)
            elif bond_loc1 == 5 and bond_loc2 == 4:  # Reverse 4O5 linkage
                write_patch(f, "4O4", chain_id, psf_patch_resid2, psf_patch_resid1)
            elif bond_loc1 == 8 and bond_loc2 == 1:  # Beta-1 linkage
                write_patch(f, "B1", chain_id, psf_patch_resid1, psf_patch_resid2)
            elif bond_loc1 == 1 and bond_loc2 == 8:  # Reverse beta-1 linkage
                write_patch(f, "B1", chain_id, psf_patch_resid2, psf_patch_resid1)
            elif bond_loc1 == -8 and bond_loc2 == 4:  # Beta-1 linkage remnant
                write_patch(f, "O4AL", chain_id, psf_patch_resid2)
            elif bond_loc2 == -8 and bond_loc1 == 4:  # Reverse beta-1 remnant
                write_patch(f, "O4AL", chain_id, psf_patch_resid1)
            elif bond_loc1 == -8 and bond_loc2 == 1:  # Beta-1 linkage remnant (C1 variant)
                write_patch(f, "C1AL", chain_id, psf_patch_resid2)
            elif bond_loc2 == -8 and bond_loc1 == 1:  # Reverse beta-1 remnant (C1 variant)
                write_patch(f, "C1AL", chain_id, psf_patch_resid1)
            elif bond_loc1 == 8 and bond_loc2 == 8:  # beta-beta linkage
                write_patch(f, "BB", chain_id, psf_patch_resid1, psf_patch_resid2)
            else:
                raise InvalidDataError(f"Encountered unexpected linkage: adj_matrix loc: {bond_matrix_tuple}, "
                                       f"bond locations: {bond_loc1} and {bond_loc2}, monomer types: "
                                       f"{monomers[bond_matrix_tuple[0]].type} and "
                                       f"{monomers[bond_matrix_tuple[1]].type}")
        f.write(f"regenerate angles dihedrals\nwritepsf {psf_fname}.psf\n")



# !/usr/bin/env python
# coding=utf-8

from common_wrangler.common import InvalidDataError

# lignin subunits (phenylpropanoids)
G = 'guaiacol'  # (coniferyl alcohol)
S = 'syringyl'  # (sinapyl alcohol)
C = 'caffeoyl'  # (caffeyl alcohol)
H = 'p-hydroxyphenyl'
S4 = 'S4'
G4 = 'G4'
G7 = 'G7'
LIGNIN_SUBUNITS = [G, S, H, C]
# Dict below likely to be changed when H added
INT_TO_TYPE_DICT = {0: G, 1: S}
MONOLIG_OHS = {G: 'coniferyl', S: 'sinapyl', H: 'p-coumaryl', C: 'caffeoyl'}
ADD_RATE = 'add_rate'
INI_MONOS = 'initial_num_monomers'
MAX_MONOS = 'max_num_monomers'
SIM_TIME = 'length_simulation'
RANDOM_SEED = 'random_seed'
CHAIN_ID = 'chain_id'
PSF_FNAME = 'psf_fname'
TOPPAR_DIR = "toppar_dir"
# specifying a random seed is not enough for reliable testing, as a generated float can differ by machine precision
# Thus, MAX_NUM_DECIMAL is used to round floats to lower-than machine precision
MAX_NUM_DECIMAL = 8

# bond types
AO4 = 'ao4'
B1 = 'b1'
B1_ALT = 'b1alt'
B5 = 'b5'
BB = 'bb'
BO4 = 'bo4'
C5C5 = '55'
C5O4 = '5o4'
BOND_TYPE_LIST = [BO4, BB, B5, B1, C5O4, AO4, C5C5]

# reaction types other than bond formation
Q = 'hydration'
OX = 'oxidation'
GROW = 'grow'

OLIGOMER = 'oligomer'
MONOMER = 'monomer'
MON_MON = (MONOMER, MONOMER)
MON_OLI = (MONOMER, OLIGOMER)
OLI_MON = (OLIGOMER, MONOMER)
OLI_OLI = (OLIGOMER, OLIGOMER)

TIME = 'time'
TEMP = 'temperature_in_k'
E_BARRIER_KCAL_MOL = 'e_barrier_in_kcal_mol'
E_BARRIER_J_PART = 'e_barrier_in_j_particle'

AFFECTED = 'affected'
ADJ_MATRIX = 'adjacency_matrix'
MONO_LIST = 'monomers'

ATOMS = 'Atoms'
BONDS = 'Bonds'
CHAIN_LEN = 'Chain Lengths'
CHAIN_MONOS = 'Chain Monomers'
CHAIN_BRANCHES = 'Chain Branches'
CHAIN_BRANCH_COEFF = 'Chain Branch Coefficients'
RCF_BONDS = 'RCF Bonds'
RCF_YIELDS = 'RCF Yields'
RCF_MONOS = 'RCF Monomers'
RCF_BRANCHES = 'RCF Branches'
RCF_BRANCH_COEFF = 'RCF Branch Coefficients'

DEF_TCL_FNAME = "psfgen.tcl"
DEF_CHAIN_ID = 'L'
DEF_PSF_FNAME = 'lignin'
DEF_TOPPAR = "toppar/"

# Data

# Default Gibbs free energy barriers input in kcal/mol (at 298.15 K and 1 atm) from Gani et al., ACS Sustainable Chem.
#     Eng. 2019, 7, 15, 13270-13277, https://doi.org/10.1021/acssuschemeng.9b02506, as described in Orella et al., ACS
#     Sustainable Chem. Eng. 2019, https://doi.org/10.1021/acssuschemeng.9b03534
# Per Terry Gani: the solution state correction is not needed because this barrier is based on the TS vs. the
#     the reactant (hydrogen-bonded) complex
DEF_E_BARRIER_KCAL_MOL = {C5O4: {(G, G): {(MONOMER, MONOMER): 11.2, (MONOMER, OLIGOMER): 14.6,
                                          (OLIGOMER, MONOMER): 14.6, (OLIGOMER, OLIGOMER): 4.4},
                                 (S, G): {(MONOMER, MONOMER): 10.9, (MONOMER, OLIGOMER): 14.6,
                                          (OLIGOMER, MONOMER): 14.6, (OLIGOMER, OLIGOMER): 4.4},
                                 (C, C): {(MONOMER, MONOMER): 11.9, (MONOMER, OLIGOMER): 11.9,
                                          (OLIGOMER, MONOMER): 11.9, (OLIGOMER, OLIGOMER): 11.9}},
                          C5C5: {(G, G): {(MONOMER, MONOMER): 12.5, (MONOMER, OLIGOMER): 15.6,
                                          (OLIGOMER, MONOMER): 15.6, (OLIGOMER, OLIGOMER): 3.8},
                                 (C, C): {(MONOMER, MONOMER): 10.6, (MONOMER, OLIGOMER): 10.6,
                                          (OLIGOMER, MONOMER): 10.6, (OLIGOMER, OLIGOMER): 10.6}},
                          B5: {(G, G): {(MONOMER, MONOMER): 5.5, (MONOMER, OLIGOMER): 5.8, (OLIGOMER, MONOMER): 5.8,
                                        (OLIGOMER, OLIGOMER): 5.8},
                               (G, S): {(MONOMER, MONOMER): 5.5, (MONOMER, OLIGOMER): 5.8, (OLIGOMER, MONOMER): 5.8,
                                        (OLIGOMER, OLIGOMER): 5.8},
                               (C, C): {(MONOMER, MONOMER): 1.9, (MONOMER, OLIGOMER): 5.8, (OLIGOMER, MONOMER): 5.8,
                                        (OLIGOMER, OLIGOMER): 5.8}},
                          BB: {(G, G): {(MONOMER, MONOMER): 5.2, (MONOMER, OLIGOMER): 5.2, (OLIGOMER, MONOMER): 5.2,
                                        (OLIGOMER, OLIGOMER): 5.2},
                               (S, G): {(MONOMER, MONOMER): 6.5, (MONOMER, OLIGOMER): 6.5, (OLIGOMER, MONOMER): 6.5,
                                        (OLIGOMER, OLIGOMER): 6.5},
                               (G, S): {(MONOMER, MONOMER): 6.5, (MONOMER, OLIGOMER): 6.5, (OLIGOMER, MONOMER): 6.5,
                                        (OLIGOMER, OLIGOMER): 6.5},
                               (S, S): {(MONOMER, MONOMER): 5.2, (MONOMER, OLIGOMER): 5.2, (OLIGOMER, MONOMER): 5.2,
                                        (OLIGOMER, OLIGOMER): 5.2},
                               (C, C): {(MONOMER, MONOMER): 7.2, (MONOMER, OLIGOMER): 7.2, (OLIGOMER, MONOMER): 7.2,
                                        (OLIGOMER, OLIGOMER): 7.2}},
                          BO4: {(G, G): {(MONOMER, MONOMER): 6.3, (MONOMER, OLIGOMER): 6.2, (OLIGOMER, MONOMER): 6.2,
                                         (OLIGOMER, OLIGOMER): 6.2},
                                (S, G): {(MONOMER, MONOMER): 9.1, (MONOMER, OLIGOMER): 6.2,
                                         (OLIGOMER, MONOMER): 6.2, (OLIGOMER, OLIGOMER): 6.2},
                                (G, S): {(MONOMER, MONOMER): 8.9, (MONOMER, OLIGOMER): 6.2,
                                         (OLIGOMER, MONOMER): 6.2, (OLIGOMER, OLIGOMER): 6.2},
                                (S, S): {(MONOMER, MONOMER): 9.8, (MONOMER, OLIGOMER): 10.4,
                                         (OLIGOMER, MONOMER): 10.4, (OLIGOMER, OLIGOMER): 10.4},
                                (C, C): {(MONOMER, MONOMER): 4.9, (MONOMER, OLIGOMER): 1.3,
                                         (OLIGOMER, MONOMER): 1.3, (OLIGOMER, OLIGOMER): 1.3}},
                          AO4: {(G, G): {(MONOMER, MONOMER): 20.7, (MONOMER, OLIGOMER): 20.7,
                                         (OLIGOMER, MONOMER): 20.7, (OLIGOMER, OLIGOMER): 20.7},
                                (S, G): {(MONOMER, MONOMER): 20.7, (MONOMER, OLIGOMER): 20.7,
                                         (OLIGOMER, MONOMER): 20.7, (OLIGOMER, OLIGOMER): 20.7},
                                (G, S): {(MONOMER, MONOMER): 20.7, (MONOMER, OLIGOMER): 20.7,
                                         (OLIGOMER, MONOMER): 20.7, (OLIGOMER, OLIGOMER): 20.7},
                                (S, S): {(MONOMER, MONOMER): 20.7, (MONOMER, OLIGOMER): 20.7,
                                         (OLIGOMER, MONOMER): 20.7, (OLIGOMER, OLIGOMER): 20.7},
                                (C, C): {(MONOMER, MONOMER): 20.7, (MONOMER, OLIGOMER): 20.7,
                                         (OLIGOMER, MONOMER): 20.7, (OLIGOMER, OLIGOMER): 20.7}},
                          B1: {(G, G): {(MONOMER, OLIGOMER): 9.6, (OLIGOMER, MONOMER): 9.6,
                                        (OLIGOMER, OLIGOMER): 9.6},
                               (S, G): {(MONOMER, OLIGOMER): 11.7, (OLIGOMER, MONOMER): 11.7,
                                        (OLIGOMER, OLIGOMER): 11.7},
                               (G, S): {(MONOMER, OLIGOMER): 10.7, (OLIGOMER, MONOMER): 10.7,
                                        (OLIGOMER, OLIGOMER): 10.7},
                               (S, S): {(MONOMER, OLIGOMER): 11.9, (OLIGOMER, MONOMER): 11.9,
                                        (OLIGOMER, OLIGOMER): 11.9},
                               (C, C): {(MONOMER, OLIGOMER): 9.6, (OLIGOMER, MONOMER): 9.6,
                                        (OLIGOMER, OLIGOMER): 9.6}},
                          OX: {G: {MONOMER: 0.9, OLIGOMER: 6.3}, S: {MONOMER: 0.6, OLIGOMER: 2.2},
                               C: {MONOMER: 0.9, OLIGOMER: 0.9}},
                          Q: {G: {MONOMER: 11.1, OLIGOMER: 11.1}, S: {MONOMER: 11.7, OLIGOMER: 11.7},
                              C: {MONOMER: 11.1, OLIGOMER: 11.1}}}

# These were calculated at 298 K from the DEF_E_BARRIER_KCAL_MOL
DEF_RXN_RATES = {C5O4: {(G, G): {MON_MON: 38335.5972148372, MON_OLI: 123.419593715543, OLI_MON: 123.419593715543,
                                 OLI_OLI: 3698609451.84164},
                        (S, G): {MON_MON: 63606.8417529500, MON_OLI: 123.419593715543, OLI_MON: 123.419593715543,
                                 OLI_OLI: 3698609451.84164},
                        (C, C): {MON_MON: 11762.4692901771, MON_OLI: 11762.4692901771, OLI_MON: 11762.4692901771,
                                 OLI_OLI: 11762.4692901771}},
                 C5C5: {(G, G): {MON_MON: 4272.63018912086, MON_OLI: 22.8233180720356, OLI_MON: 22.8233180720356,
                                 OLI_OLI: 10182201166.0217},
                        (C, C): {MON_MON: 105537.166803781, MON_OLI: 105537.166803781, OLI_MON: 105537.166803781,
                                 OLI_OLI: 105537.166803781}},
                 B5: {(G, G): {MON_MON: 577740233.381881, MON_OLI: 348201801.431315, OLI_MON: 348201801.431315,
                               OLI_OLI: 348201801.431315},
                      (G, S): {MON_MON: 577740233.381881, MON_OLI: 348201801.431315, OLI_MON: 348201801.431315,
                               OLI_OLI: 348201801.431315},
                      (C, C): {MON_MON: 251507997491.634, MON_OLI: 348201801.431315, OLI_MON: 348201801.431315,
                               OLI_OLI: 348201801.431315}},
                 BB: {(G, G): {MON_MON: 958592907.607318, MON_OLI: 958592907.607318, OLI_MON: 958592907.607318,
                               OLI_OLI: 958592907.607318},
                      (S, G): {MON_MON: 106838377.218107, MON_OLI: 106838377.218107, OLI_MON: 106838377.218107,
                               OLI_OLI: 106838377.218107},
                      (G, S): {MON_MON: 106838377.218107, MON_OLI: 106838377.218107, OLI_MON: 106838377.218107,
                               OLI_OLI: 106838377.218107},
                      (S, S): {MON_MON: 958592907.607318, MON_OLI: 958592907.607318, OLI_MON: 958592907.607318,
                               OLI_OLI: 958592907.607318},
                      (C, C): {MON_MON: 32781102.2219828, MON_OLI: 32781102.2219828, OLI_MON: 32781102.2219828,
                               OLI_OLI: 32781102.2219828}},
                 BO4: {(G, G): {MON_MON: 149736731.431189, MON_OLI: 177267402.79460, OLI_MON: 177267402.794600,
                                OLI_OLI: 177267402.794600},
                       (S, G): {MON_MON: 1327129.87498242, MON_OLI: 177267402.79460, OLI_MON: 177267402.794600,
                                OLI_OLI: 177267402.794600},
                       (G, S): {MON_MON: 1860006.62719604, MON_OLI: 177267402.79460, OLI_MON: 177267402.794600,
                                OLI_OLI: 177267402.794600},
                       (S, S): {MON_MON: 407201.805441432, MON_OLI: 147913.051594236, OLI_MON: 147913.051594236,
                                OLI_OLI: 147913.051594236},
                       (C, C): {MON_MON: 1590507825.87210, MON_OLI: 692396712512.577, OLI_MON: 692396712512.577,
                                OLI_OLI: 692396712512.577}},
                 AO4: {(G, G): {MON_MON: 0.00416918917397265, MON_OLI: 0.00416918917397265,
                                OLI_MON: 0.00416918917397265, OLI_OLI: 0.00416918917397265},
                       (S, G): {MON_MON: 0.00416918917397265, MON_OLI: 0.00416918917397265,
                                OLI_MON: 0.00416918917397265, OLI_OLI: 0.00416918917397265},
                       (G, S): {MON_MON: 0.00416918917397265, MON_OLI: 0.00416918917397265,
                                OLI_MON: 0.00416918917397265, OLI_OLI: 0.00416918917397265},
                       (S, S): {MON_MON: 0.00416918917397265, MON_OLI: 0.00416918917397265,
                                OLI_MON: 0.00416918917397265, OLI_OLI: 0.00416918917397265},
                       (C, C): {MON_MON: 0.00416918917397265, MON_OLI: 0.00416918917397265,
                                OLI_MON: 0.00416918917397265, OLI_OLI: 0.00416918917397265}},
                 B1: {(G, G): {MON_OLI: 570703.795464849, OLI_MON: 570703.795464849, OLI_OLI: 570703.795464849},
                      (S, G): {MON_OLI: 16485.4030071542, OLI_MON: 16485.4030071542, OLI_OLI: 16485.4030071542},
                      (G, S): {MON_OLI: 89146.6234207596, OLI_MON: 89146.6234207596, OLI_OLI: 89146.6234207596},
                      (S, S): {MON_OLI: 11762.4692901771, OLI_MON: 11762.4692901771, OLI_OLI: 11762.4692901771},
                      (C, C): {MON_OLI: 570703.795464849, OLI_MON: 570703.795464849, OLI_OLI: 570703.795464849}},
                 OX: {G: {MONOMER: 1360057059567.54, OLIGOMER: 149736731.431189},
                      S: {MONOMER: 2256621533195.09, OLIGOMER: 151582896154.443},
                      C: {MONOMER: 1360057059567.54, OLIGOMER: 1360057059567.54}},
                 Q: {G: {MONOMER: 45383.9995564285, OLIGOMER: 45383.9995564285},
                     S: {MONOMER: 16485.4030071542, OLIGOMER: 16485.4030071542},
                     C: {MONOMER: 45383.9995564285, OLIGOMER: 45383.9995564285}}
                 }

# todo: remove MANUSCRIPT_RATES when debugging done
MANUSCRIPT_RATES = {C5O4: {(G, G): {MON_MON: 5904261.55598695, MON_OLI: 80333.560184945, OLI_MON: 80333.560184945,
                                    OLI_OLI: 31893540751.2937},
                           (S, G): {MON_MON: 8626534.12830123, MON_OLI: 80333.560184945, OLI_MON: 80333.560184945,
                                    OLI_OLI: 31893540751.2937}},
                    C5C5: {(G, G): {MON_MON: 1141805.97148106, MON_OLI: 22698.3606666981, OLI_MON: 22698.3606666981,
                                    OLI_OLI: 68083872447.6958}},
                    B5: {(G, G): {MON_MON: 7941635722.59467, MON_OLI: 5435496317.66216, OLI_MON: 5435496317.66216,
                                  OLI_OLI: 5435496317.66216},
                         (G, S): {MON_MON: 7941635722.59467, MON_OLI: 5435496317.66216, OLI_MON: 5435496317.66216,
                                  OLI_OLI: 5435496317.66216}},
                    BB: {(G, G): {MON_MON: 11603278571.9039, MON_OLI: 11603278571.9039, OLI_MON: 11603278571.9039,
                                  OLI_OLI: 11603278571.9039},
                         (S, G): {MON_MON: 2243920367.77638, MON_OLI: 2243920367.77638, OLI_MON: 2243920367.77638,
                                  OLI_OLI: 2243920367.77638},
                         (S, S): {MON_MON: 11603278571.9039, MON_OLI: 11603278571.9039, OLI_MON: 11603278571.9039,
                                  OLI_OLI: 11603278571.9039},
                         (G, S): {MON_MON: 2243920367.77638, MON_OLI: 2243920367.77638, OLI_MON: 2243920367.77638,
                                  OLI_OLI: 2243920367.77638}},
                    BO4: {(G, G): {MON_MON: 2889268780.92427, MON_OLI: 3278522716.22094, OLI_MON: 3278522716.22094,
                                   OLI_OLI: 3278522716.22094},
                          (S, G): {MON_MON: 83919112.8376677, MON_OLI: 3278522716.22094, OLI_MON: 3278522716.22094,
                                   OLI_OLI: 3278522716.22094},
                          (G, S): {MON_MON: 108054134.329644, MON_OLI: 3278522716.22094, OLI_MON: 3278522716.22094,
                                   OLI_OLI: 3278522716.22094},
                          # below is where an entry is missing
                          (S, S): {MON_MON: 34644086.8574001, MON_OLI: 16228844.7506668, OLI_MON: 16228844.7506668}},
                    AO4: {(G, G): {MON_MON: 36.0239749057507, MON_OLI: 36.0239749057507, OLI_MON: 36.0239749057507,
                                   OLI_OLI: 36.0239749057507},
                          (S, G): {MON_MON: 36.0239749057507, MON_OLI: 36.0239749057507, OLI_MON: 36.0239749057507,
                                   OLI_OLI: 36.0239749057507},
                          (G, S): {MON_MON: 36.0239749057507, MON_OLI: 36.0239749057507, OLI_MON: 36.0239749057507,
                                   OLI_OLI: 36.0239749057507},
                          (S, S): {MON_MON: 36.0239749057507, MON_OLI: 36.0239749057507, OLI_MON: 36.0239749057507,
                                   OLI_OLI: 36.0239749057507}},
                    B1: {(G, G): {MON_OLI: 44607678.613794, OLI_MON: 44607678.613794, OLI_OLI: 44607678.613794},
                         (S, G): {MON_OLI: 3138443.59211371, OLI_MON: 3138443.59211371, OLI_OLI: 3138443.59211371},
                         (G, S): {MON_OLI: 11107513.4850607, OLI_MON: 11107513.4850607, OLI_OLI: 11107513.4850607},
                         (S, S): {MON_OLI: 2437439.37772669, OLI_MON: 2437439.37772669, OLI_OLI: 2437439.37772669}},
                    OX: {G: {MONOMER: 2659877051606.15, OLIGOMER: 2889268780.92427},
                         S: {MONOMER: 3886264174644.99, OLIGOMER: 514384986527.191}},
                    Q: {G: {MONOMER: 6699707.46979824, OLIGOMER: 6699707.46979824},
                        S: {MONOMER: 3138443.59211371, OLIGOMER: 3138443.59211371}}}


class Event:
    """
    Class definition for the event_dict that occur during lignification process. While more specific than the monomer
    class, this class is also easily extensible such that other reactivity could be incorporated in the future. The
    most important features lie in the event dictionary that specifies the changes that need to occur for a given
    reaction.

    ATTRIBUTES:
        key     -- str  --  Name of the bond that is being formed using the traditional nomenclature
                            (e.g. C5o4, BO4, C5C5 etc.)
        index   -- list --  N x 1 list of integers containing the unique monomer identifiers involved in the reaction
                            (N = 2 for bimolecular, 1 for unimolecular)
        rate    -- float--  Scalar floating point value with the rate of this particular reaction event
        bond    -- list --  2 x 1 list of the updates that need to occur to the adjacency matrix to reflect the
                            bond formations

    METHODS:
        N/A no declared public methods

    Events are compared based on the monomers that are involved in the event, the specific bond being formed, and the
    value updates to the adjacency matrix.
    """
    # create a dict that maps event keys onto numerical changes in monomer
    # state where the value is a tuple of (new reactant active point, openPos0, openPos1)
    eventDict = {BO4: ((-1, 7), (), (7,)), BB: ((0, 0), (), ()),
                 C5C5: ((0, 0), (), ()), C5O4: ((-1, 0), (), ()),
                 B5: ((-1, 0), (), ()), B1: ((0, 7), (), (7,)),
                 AO4: ((-1, 4), (), ()), Q: (0, (1,), ()), OX: (4, (), ())}

    activeDict = {(4, 8): (0, 1), (8, 4): (1, 0), (4, 5): (0, 1), (5, 4): (1, 0), (5, 8): (0, 1), (8, 5): (1, 0),
                  (1, 8): (0, 1), (8, 1): (1, 0), (4, 7): (0, 1), (7, 4): (1, 0), (5, 5): (0, 1), (8, 8): (0, 1)}

    def __init__(self, event_name, ids, rate=0, bond=()):
        """
        Straightforward constructor that takes the name of bond being formed, the indices of affected monomers, the
        rate of the event, and the updates to the adjacency matrix.

        :param event_name: str, assigned to key attribute - has the name of the bond being formed or the reaction
                               occurring
        :param ids:  int, N x 1 list assigned to index - keeps track of the monomers affected by this event
        :param rate: float, the rate of the event (make sure units are consistent - I typically use GHz)
        :param bond: int, 2 x 1 list of updates to the adjacency matrix when a bond is formed
        :return: new instance of an event object that corresponds to the attributes provided
        """
        self.key = event_name
        self.index = ids
        self.rate = rate
        self.bond = bond

    def __str__(self):
        if self.key == Q or self.key == OX:
            msg = f'Performing {self.key} on index {str(self.index[0])}'
        else:
            msg = f'Forming {self.key} bond between indices {str(self.index)} ({ADJ_MATRIX} update {str(self.bond)})'
        return msg

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.index == other.index and self.bond == other.bond and self.key == other.key

    def __lt__(self, other):
        return self.index < other.index

    def __hash__(self):
        # Note: changed from invoking python's hash function to provide more consistent output for testing
        #     see https://docs.python.org/3/reference/datamodel.html#object.__hash__
        #     "By default, the __hash__() values of str and bytes objects are “salted” with an unpredictable random
        #     value. Although they remain constant within an individual Python process, they are not predictable
        #     between repeated invocations of Python."
        key_as_num = sum([ord(x) % 32 for x in self.key])
        return key_as_num + sum(self.index) * 1000 + int(self.rate * 10000)


class Monomer:
    """
    Class definition for monomer objects. This is highly generic such that it can be easily extended to more types of
    monolignols. The class is primarily used for storing information about each monolignol included in a simulation of
    polymerization.

    ATTRIBUTES:
        identity    -- int     --  unique integer for indexing monomer (also the hash value)
        type        -- str     --  monolignol variety
                                        G = coniferyl alcohol
                                        S = sinapyl alcohol
                                        C = caffeoyl alcohol
        parent      -- Monomer --  monomer object that has the smallest unique identifier in a chain (can be used for
                                    sizing fragments)
        size        -- int     --  integer with the size of the fragment if parent == self
        active      -- int     --  integer with the location [1-9] of the active site on the monomer (-1 means
                                    inactive)
        open        -- set     --  set of units with location [1-9] of open positions on the monomer
        connectedTo -- set     --  set of integer identities of other monomers that this monomer is connected to

    METHODS:
        N/A - no defined public methods

    Monomers are mutable objects, but compare and hash only on the identity, which should be treated as a constant
    """

    def __init__(self, unit_type, i):
        """
        Constructor for the Monomer class, which sets the properties depending on what monolignol is being represented.
        The only attributes that need to be set are the species [0-2], and the unique integer identifier. Everything
        else will be computed from these values. The active site is initially set to 0, indicating that the monomer is
        not oxidized, but can be eventually. The open positions are either {4,5,8} or {4,8} depending on whether a
        5-methoxy is present. The parent is set to be self until connections occur. The set of monomers that are
        connected to begin as just containing the self's identity.

        Example calls are below:
            mon = Monomer(G, 0) # Makes a guaiacol unit monomer with ID = 0
            mon = Monomer(S, 0) # Makes a syringol unit monomer with ID = 0 (not recommended to repeat IDs)
            mon = Monomer(H, 0) # Makes a caffeoyl unit monomer with ID = 0
            mon = Monomer(S, n) # Makes a sinapyl alcohol with ID = n

        :param unit_type: str, monomer type
        :param i: int, unique identifier for the monomer
        Outputs:
            New instance of a monomer object with the desired attributes
        """

        self.identity = i
        self.type = unit_type
        self.parent = self
        self.size = 1

        # The active attribute will be the position of an active position, if 0
        # the monomer is not activated yet, -1 means it can never be activated
        self.active = 0
        if unit_type == G or unit_type == C:
            self.open = {4, 5, 8}
        elif unit_type == S:
            self.open = {4, 8}
        else:
            # todo: update once H is added
            raise InvalidDataError(f"Encountered unit type {unit_type},  but only the following types are "
                                   f"currently available: 'G' ({G}), 'S' ({S}), 'C' ({C})")
        self.connectedTo = {i}

    def __str__(self):
        return f'{self.identity}: {MONOLIG_OHS[self.type]} alcohol is connected to {self.connectedTo} and active at ' \
               f'position {self.active}'

    def __repr__(self):
        representation = f'{self.identity}: {MONOLIG_OHS[self.type]} alcohol \n'
        return representation

    def __eq__(self, other):  # Always compare monomers by identity alone. This should always be a unique identifier
        return self.identity == other.identity

    def __lt__(self, other):
        return self.identity < other.identity

    def __hash__(self):
        return self.identity



# -*- coding: utf-8 -*-
"""
ligninkmc
Kinetic Monte Carlo implementation for creating realistic lignin topologies as described in
    https://pubs.acs.org/doi/abs/10.1021/acssuschemeng.9b03534
"""

from setuptools import setup
import versioneer

DOCLINES = __doc__.split("\n")

setup(name='ligninkmc',
      author='Michael Orella, Terry Gani, and Heather Mayes',
      author_email='hmayes@gmail.com',
      # description="Kinetic Monte Carlo implementation for creating realistic lignin topologies.",
      description=DOCLINES[0],
      # long_description="\n".join(DOCLINES[2:]),
      url='https://github.com/michaelorella/lignin-kmc',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      license='MIT',
      packages=['ligninkmc'],
      entry_points={'console_scripts': ['create_lignin = ligninkmc.create_lignin:main',
                                        'plot_bond_formation = ligninkmc.plot_bond_formation:main'
                                        ],
                    },     package_dir={'ligninkmc': 'ligninkmc'},
      python_requires=">3.6",  # Required for f-strings support
      test_suite='tests',
      # rdkit isn't installable via pip, so not included on list below
      install_requires=['scipy', 'numpy', 'matplotlib', 'common_wrangler>=0.2.3'],
      # zip_safe=False,
      )
